/*  $Id: bootmanager.cpp,v 1.1 2001/01/25 14:12:56 jan Exp $

    Part of SWI-Prolog CORBA Interface
    Designed and implemented by Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 2001 University of Amsterdam. All rights reserved.
*/

#include <OB/CORBA.h>
#include <OB/OCI_IIOP.h>
#include <OB/BootManager.h>
#include <SWI-cpp.h>
#include <pl2ilu.h>

static
OB::BootManager_ptr
findBootManager()
{ static OB::BootManager_ptr mgr = NULL;

  if ( !mgr )
  { CORBA::ORB_ptr orb = PL_CORBA_default_server();
    CORBA::Object_ptr obj = orb->resolve_initial_references("BootManager");
    mgr = OB::BootManager::_narrow(obj);
  }

  return mgr;
}


PREDICATE(corba_bind, 2)
{ void *ptr = A1;
  CORBA::Object_ptr obj = (CORBA::Object_ptr)ptr;
  OB::BootManager_ptr mgr = findBootManager();

  PortableServer::ObjectId_var id = PortableServer::string_to_ObjectId(A2);
  mgr->add_binding(id, obj);

  return TRUE;
}


PREDICATE(corba_iiop_info, 1)
{ PlTail tail = A1;

  CORBA::ORB_ptr orb = PL_CORBA_default_server();
  CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
  PortableServer::POA_var root = PortableServer::POA::_narrow(obj);
  PortableServer::POAManager_var poaManager = root -> the_POAManager();
  OBPortableServer::POAManager_var obManager =
		    OBPortableServer::POAManager::_narrow(poaManager);

  OCI::AcceptorSeq_var acceptors = obManager->get_acceptors();

  for(CORBA::ULong i = 0; i < acceptors->length() ; i++)
  { OCI::AcceptorInfo_var info = acceptors[i]->get_info();
    OCI::IIOP::AcceptorInfo_var iiopInfo =
	       OCI::IIOP::AcceptorInfo::_narrow(info);

    if ( !CORBA::is_nil(iiopInfo) )
    { CORBA::StringSeq_var hosts = iiopInfo -> hosts();
      CORBA::UShort port = iiopInfo -> port();

      tail.append(PlCompound(":", PlTermv((char *)hosts[0], (long)port)));
    }
  }
  tail.close();

  return TRUE;
}
