/*  $Id: client.pl,v 1.2 2001/01/25 14:12:56 jan Exp $

    Part of SWI-Prolog CORBA interface
    Designed and implemented by Anjo Anjewierden and Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1998 University of Amsterdam. All rights reserved.
*/

:- use_module(tests).

init :-
	load_foreign_library(client, 'PL_test__Initialize'),
	corba_initialize_orb([], _).

:- initialization init.

%	test

test :-
	test(localhost:8000).

test(Port) :-
	test_interface(Name),
	interface(Port, Name, Interface),
	run_interface_tests(Name, Interface),
	nl.
test(_).

run_interface_tests(input, Interface) :- !,
	test_data(Data, Term),
	concat_atom([input, t, Data], '_', Pred),
	progress(call(Pred, Interface, Term)),
	fail.
run_interface_tests(Name, Interface) :-
	memberchk(Name, [answer, output]),
	test_data(Data, Term),
	concat_atom([Name, t, Data], '_', Pred),
	format('~N~w ... ', [Pred]), flush,
	call(Pred, Interface, Rval),
	(   Rval == Term
	->  format('ok~n', [])
	;   format('Wrong answer: ~w~n', [Rval])
	),
	fail.
run_interface_tests(except, Interface) :- !,
	test_data(Data, Term),
	concat_atom([except, t, Data], '_', Pred),
	format('~N~w ... ', [Pred]), flush,
	catch(call(Pred, Interface), E, true),
	concat_atom([t_except, Data], '_', Functor),
	ETerm =.. [Functor, Term],
	(   E == ETerm
	->  format('ok~n', [])
	;   format('Wrong answer: ~w!~n', [E])
	),
	fail.

progress(Goal) :-
	progress_message(Goal),
	(   Goal
	->  format('ok~n', [])
	;   format('FAILED~n', [])
	).
	
progress_message(call(Pred, _, _)) :-
	format('~N~w ... ', [Pred]),
	flush.


		 /*******************************
		 *	     INTERFACES		*
		 *******************************/

:- dynamic
	interface_cache/2.

interface(_, Name, Interface) :-
	interface_cache(Name, Interface), !.
interface(Port, Name, Interface) :-
	fabriek(Port, Factory),
	concat_atom([fabriek_make, Name], '_', Pred),
	call(Pred, Factory, Interface),
	asserta(interface_cache(Name, Interface)).


		 /*******************************
		 *	     FACTORY		*
		 *******************************/

:- dynamic
	'_test_fabriek'/1.

fabriek(_Port, Ref) :-
	'_test_fabriek'(Ref), !.
fabriek(Port, Ref) :-
	url(Port, URL),
	corba_string_to_object(URL, Ref),
	asserta('_test_fabriek'(Ref)).

url(Host:Port, URL) :- !,
	sformat(URL, 'corbaloc::~w:~w/fabriek', [Host, Port]).
url(Port, URL) :- !,
	sformat(URL, 'corbaloc::localhost:~w/fabriek', [Port]).
