/*  $Id: cosnaming.cpp,v 1.1.1.1 2001/01/24 15:20:58 jan Exp $

    Part of XPCE
    Designed and implemented by Anjo Anjewierden and Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 2001 University of Amsterdam. All rights reserved.
*/

#include <SWI-cpp.h>
#include <OB/CORBA.h>
#include <OB/CosNaming.h>
#include <pl2ilu.h>

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Provide access to the CORBA CosNaming  service. This module provides the
following primitives:

Server:

	corba_bind(Object, Name)
	corba_unbind(Name)


Client:
	corba_resolve(Name, Object)
	
Where `Name' is a path of the   form N1/N2/.. and each name-component is
either a single atom or a term of the form kind(Name).
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static int
NameService(CosNaming::NamingContext_ptr &nc)
{ CORBA::ORB_ptr orb = PL_CORBA_default_server();
  CORBA::Object_var obj;

  try
  { orb->resolve_initial_references("NameService");
  } catch (const CORBA::ORB::InvalidName&)
  { return PL_warning("Couldn't resolve NameService");
  }
  if ( CORBA::is_nil(obj.in()) )
    return PL_warning("NameService is nil");

  nc = CosNaming::NamingContext::_narrow(obj.in());
  if ( CORBA::is_nil(nc) )
    return PL_warning("NameService is not a NamingContext object");

  return TRUE;
}


static int
is_div(PlTerm in)
{ return in.type() == PL_TERM && in.name() == "/" && in.arity() == 2;
}


static void
set_nc(CosNaming::Name& n, int pos, PlTerm t)
{ if ( t.type() == PL_ATOM )
  { n[pos].id   = CORBA::string_dup(t);
    n[pos].kind = CORBA::string_dup("");
  } else				/* TBD: test type */
  { n[pos].id   = CORBA::string_dup(t[1]);
    n[pos].kind = CORBA::string_dup(t.name());
  }
}


static void
get_name(PlTerm in, CosNaming::Name& n)
{ int len = 1;
  PlTerm tmp = in;

  while( is_div(tmp) )
  { tmp = tmp[1];
    len++;
  }

  n.length(len);

  PlTerm t2 = in;
  while( is_div(t2) )
  { set_nc(n, --len, t2[2]);
    t2 = t2[1];
  }
  set_nc(n, 0, t2);
}


PREDICATE(corba_bind, 2)
{ CosNaming::NamingContext_ptr nc;
  if ( !NameService(nc) )
    return FALSE;

  void *ptr = A1;
  CORBA::Object_ptr obj = (CORBA::Object_ptr)ptr;

  CosNaming::Name name;
  get_name(A2, name);

  nc->bind(name, obj);
  return TRUE;
}


PREDICATE(corba_unbind, 1)
{ CosNaming::NamingContext_ptr nc;
  if ( !NameService(nc) )
    return FALSE;

  CosNaming::Name name;
  get_name(A1, name);

  nc->unbind(name);
}


PREDICATE(corba_resolve, 2)
{ CosNaming::NamingContext_ptr nc;
  if ( !NameService(nc) )
    return FALSE;

  CosNaming::Name name;
  get_name(A1, name);

  CORBA::Object_ptr obj = nc->resolve(name);
  if ( A2 = (void *)obj )
    return TRUE;

  return FALSE;
}





