/*  $Id: mkidl.pl,v 1.1.1.1 2001/01/24 15:20:58 jan Exp $

    Part of SWI-Prolog CORBA interface
    Designed and implemented by Anjo Anjewierden and Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1998 University of Amsterdam. All rights reserved.
*/

:- module(corba_test_idl,
	  [ mkidl/0
	  ]).
:- use_module(tests).
:- use_module('pl/cutil').

%	mkidl/0
%
%	Create test.idl based on tests.pl and types.idl, defining the
%	used types.

mkidl :-
	open('test.idl', write, _Stream, [alias(cout)]),
	header,
	test_interface(Name),
	mkinterface(Name),
	fail.
mkidl :-
	footer,
	close(cout).

type(T) :-
	findall(Type, Data^test_data(Type, Data), Types0),
	list_to_set(Types0, Types),
	member(T, Types).

mkinterface(input) :-
	utter(['interface input', nl]),
	utter(begin),
	forall(type(Type),
	       utter(['void t_~w(in ~w a0);'+[Type, Type], nl])),
	utter([end, ';', nl, nl]).
mkinterface(answer) :-
	utter(['interface answer', nl]),
	utter(begin),
	forall(type(Type),
	       utter(['~w t_~w();'+[Type, Type], nl])),
	utter([end, ';', nl, nl]).
mkinterface(output) :-
	utter(['interface output', nl]),
	utter(begin),
	forall(type(Type),
	       utter(['void t_~w(out ~w a0);'+[Type, Type], nl])),
	utter([end, ';', nl, nl]).
mkinterface(except) :-
	forall(type(Type),
	       utter(['exception t_except_~w'+[Type], nl,
		      begin,
			'~w a0;'+[Type], nl,
		      end, ';', nl, nl])),
	utter(['interface except', nl]),
	utter(begin),
	forall(type(Type),
	       utter(['void t_~w() raises(t_except_~w);'+[Type, Type], nl])),
	utter([end, ';', nl, nl]).

header :-
	utter(['module test', nl, begin, nl, nl]),
	copy('types.idl').

footer :-
	utter([ 'interface fabriek', nl, % factory is reserved in orbacus
	        begin]),
	forall(test_interface(Name),
	       utter(['~w make_~w();'+[Name, Name], nl])),
	utter([ end, ';', nl,
		end, ';', nl
	      ]).

copy(File) :-
	open(File, read, In),
	copy_stream(In, cout),
	close(In).

copy_stream(In, Out) :-
	get0(In, C0),
	copy_stream(C0, In, Out).

copy_stream(-1, _, _) :- !.
copy_stream(C, In, Out) :-
	put(Out, C),
	get0(In, C2),
	copy_stream(C2, In, Out).
