/*  $Id: cclient.pl,v 1.1.1.1 2001/01/24 15:20:59 jan Exp $
    Part of Prolog <-> CORBA interface
    Designed and implemented by Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl
    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/
:- module(corba_cclient,
	  [ cclient_for_module/1	% +Module
	  ]).
:- use_module(util).
:- use_module(cutil).


cclient_for_module(module(Name, Defs)) :-
	setting(client_suffix, Suffix),
	(   setting(file_base, input)		
	->  setting(idl_file, IdlFile),
	    file_name_extension(Base, idl, IdlFile)
	;   Base = Name
	),
	atom_concat(Base, Suffix, Impl),
	open(Impl, write, CFd, [alias(cout)]),
	(   client_module(module(Name, Defs))
	->  close(CFd)
	;   close(CFd),
	    fail
	).
		 /*******************************
		 *	       MODULE		*
		 *******************************/
client_module(module(Name,
		     definition(dcls(_Consts, % TBD
				     Types,
				     Except),
				Interfaces,
				_Modules))) :- % TBD
	push(module, Name),
	reset_foreigns,
	foreign(corba_orb/1, pl_corba_orb),
	foreign(corba_string_to_object/2, pl_corba_string_to_object),
	foreign(corba_initialize_orb/2, pl_corba_initialize_orb),
	foreign(corba_release/1, pl_corba_release),
	utter('/* Module ~w */~n~n'+[Name]),
	include_files(Name),
	(   setting(namespaces, true)
	->  utter(['namespace ~w'+[Name], nl, '{', nl])
	;   true
	),
	progress(translate_types(Types)),
	progress(client_exception_handler(Except)),
	forall(member(Itf, Interfaces), progress(client_interface(Itf))),
	progress(module_initialiser(Name)),
	(   setting(namespaces, true)
	->  utter([nl, '}', nl])
	;   true
	),
	pop(module).

		 /*******************************
		 *	      INTERFACE		*
		 *******************************/
client_interface(interface(Name,
			   Scope,
			   _dcls,	% TBD
			   _Attrs,	% TBD
			   Methods)) :-
	(   setting(orb, orbix)
	->  progress(client_binder(Name, Scope))
	;   true
	),
%	client_destructor(Name, Scope),
	utter('/* Methods for Interface ~w */~n~n'+[Name]),
	forall(member(M, Methods),
	       progress(client_method(Name, Scope, M))).

		 /*******************************
		 *	   _BIND (ORBIX)	*
		 *******************************/
client_binder(Itf, Scope) :-
	concat_atom([Itf, '_bind'], '_', PredName),
	concat_atom([pl_corba, PredName], '_', CFunction),
	c_type(type(Itf, Scope, interface), TypeName),
	utter(['static foreign_t', nl]),
	utter(['~w(term_t self, term_t marker, term_t server, term_t host)'+
	       [CFunction], nl]),
	utter(begin),
	utter(['~w_ptr _rval;'+TypeName, nl]),
	utter(['char ms[256];', nl]),
	utter(['char *_marker = "";', nl]),
	utter(['char *_server = "";', nl]),
	utter(['char *_host = "";', nl, nl]),
	utter(['PL_get_atom_chars(host,   &_host);', nl]),
	utter(['PL_get_atom_chars(marker, &_marker);', nl]),
	utter(['PL_get_atom_chars(server, &_server);', nl, nl]),
	utter(['strcpy(ms, _marker);', nl]),
	utter(['strcat(ms, ":");', nl]),
	utter(['strcat(ms, _server);', nl, nl]),
	utter([try, nl, begin]),
	utter(['_rval = ~w::_bind(ms, _host);'+TypeName, nl]),
	utter([end, ' catch ( CORBA(SystemException) &ex ) ', nl, begin]),
	utter(['return ex_CORBA_SystemException(ex);', nl]),
	utter([end, nl, nl]),
	utter(['return PL_CORBA_unify_interface(self, _rval);', nl]),
	utter([end, nl, nl]),
	foreign(PredName/4, CFunction).

		 /*******************************
		 *	     DESTRUCTOR		*
		 *******************************/

client_destructor(Itf, Scope) :-
	setting(language, cpp), !,
	concat_atom([Itf, '_delete'], '_', PredName),
	concat_atom([pl_corba, PredName], '_', CFunction),
	c_type(type(Itf, Scope, interface), TypeName),
	utter(['static foreign_t', nl]),
	utter(['~w(term_t self)'+[CFunction], nl]),
	utter(begin),
	utter(['CORBA(Object)* _self;', nl]),
	utter(['~w_ptr obj;'+TypeName, nl, nl]),
	utter(['if ( !PL_CORBA_get_object(self, &_self, NULL) )', nl]),
	utter(['  return FALSE;', nl, nl]),
	utter(['obj = ~w::_narrow(_self);'+TypeName, nl]),
	utter(['obj->_release();', nl, nl]),
	utter(['return TRUE;', nl]),
	utter([end, nl, nl]),
	foreign(PredName/1, CFunction).
client_destructor(_, _).


		 /*******************************
		 *	      METHODS		*
		 *******************************/

client_method(Itf, Scope, method(Name, Parms, Return, Raises, _Context)) :-
	reset_local_vars,
	concat_atom([pl_corba_, Itf, '_', Name], CFunction),
	concat_atom([Itf, '_', Name], Predicate),
	arity(Parms, Return, Arity),
	utter(['static foreign_t', nl]),
	utter(['~w('+[CFunction]]),
	utter(push(indent)),
	utter(['term_t self']),
	parameters(Parms),
	return_parameter(Return),
	utter(pop(indent)),
	utter([')', nl]),
	utter(begin),
	(   setting(language, c)
	->  utter(['CORBA_Environment env;', nl])
	;   true
	),
	utter(['PL_CORBA_Environment plenv;', nl]),
	(   setting(language, cpp)
	->  utter(['CORBA(Object)* _self;', nl])
	;   utter(['void *_self;', nl])
	),
	in_declarations(Parms),
	rval_declaration(Return),
	utter(nl),
	(   setting(language, c)
	->  utter(['plenv.CORBA_env = &env;', nl])
	;   true
	),
	utter(['plenv.isserver = FALSE;', nl]),
	utter(['plenv.source.predicate_name = "~w/~d";'+
	       [Predicate, Arity], nl]), % TBD!
	utter(['if ( !PL_CORBA_get_object(self, &_self, &plenv) )', nl]),
	utter(['  return FALSE;', nl, nl]),
	read_in_parameters(Parms),
	utter(nl),
	call_corba(Itf, Scope, Name, Parms, Raises, Return),
	utter(nl),
	write_out_parameters(Parms),
	write_rval_parameter(Return),
	utter([nl, 'return TRUE;', nl]),
	utter([end, nl, nl]),
	foreign(Predicate/Arity, CFunction).


in_declarations([]).
in_declarations([in(Type, Name)|T]) :-
	c_type(Type, client(in), CType),
	utter(['~w _~w;'+[CType, Name], nl]),
	in_declarations(T).
in_declarations([out(Type, Name)|T]) :-
	c_type(Type, client(return), CType),
	utter(['~w _~w;'+[CType, Name], nl]),
	in_declarations(T).


rval_declaration(void) :- !.
rval_declaration(Type) :-
	c_type(Type, client(return), CType),
	utter(['~w _rval;'+[CType], nl]).


call_corba(Itf, Scope, Method, Parms, _Raises, Return) :-	% C-version
	setting(language, c), !,
	(   Return == void
	->  true
	;   utter('_rval = ')
	),
	flatten([Scope, Itf, Method], NameElements),
	concat_atom(NameElements, '_', FunctionName),
	utter('~w(_self, '+[FunctionName]),
	call_corba_parameters(Parms),
	utter([');', nl]),
	utter(['if ( print_exception(&env) )', nl]),
	utter(['  return FALSE;', nl]).
call_corba(Itf, Scope, Method, Parms, Raises, Return) :- 	% C++ version
	c_type(type(Itf, Scope, interface), TypeName),
	utter([try, nl, begin]),
	(   Return == void
	->  true
	;   utter('_rval = ')
	),
	utter('~w::_narrow(_self)->~w('+[TypeName, Method]),
	call_corba_parameters(Parms),
	utter([');', nl]),
	checklist(except_case, Raises),
	utter([end, ' catch ( CORBA(SystemException) &ex ) ', nl, begin]),
	utter(['return ex_CORBA_SystemException(ex);', nl]),
	utter([end, nl]).


except_case(exception(Name, Scope, _Members)) :-
	exception_type(Name, Scope, TypeName),
	exception_function(Name, Scope, FuncName),
	utter([end, ' catch ( ~w &ex ) '+[TypeName], nl, begin]),
	utter(['return ~w(ex);'+[FuncName], nl]).


call_corba_parameters([]) :-
	(   setting(language, cpp)
	->  true
	;   utter('&env')
	).
call_corba_parameters([Param|T]) :-
	call_corba_parameter(Param),
	(   T == []
	->  (   setting(language, cpp)
	    ->  true
	    ;   utter(', &env')
	    )
	;   utter(', '),
	    call_corba_parameters(T)
	).

call_corba_parameter(in(_Type, Name)) :- 
	utter('_~w'+[Name]).
call_corba_parameter(out(_Type, Name)) :- 
	setting(language, cpp),				% uses references
%	has_var_type(Type), !,
	utter('_~w'+[Name]).
call_corba_parameter(out(_Type, Name)) :- 
	utter('&_~w'+[Name]).

read_in_parameters([]).
read_in_parameters([in(Type, Name)|T]) :- 		% C-version
	setting(language, c), !,
	atom_concat('_', Name, Out),
	indirection(Type, client(in), Stars),
	(   must_be_allocated(Type)
	->  Options = [allocate, pointer(Stars)]
	;   Options = [pointer(Stars)]
	),
	prolog_read_value(Name, out(Out, Type, Options), '&plenv'),
	read_in_parameters(T).
read_in_parameters([in(Type, Name)|T]) :- !,		% C++-version
	atom_concat('_', Name, Out),
	(   must_be_allocated(Type)			% must be a _var!
	->  Options = [allocate, var]
	;   Options = []
	),
	prolog_read_value(Name, out(Out, Type, Options), '&plenv'),
	read_in_parameters(T).
read_in_parameters([out(_, _)|T]) :-
	read_in_parameters(T).

write_out_parameters([]).
write_out_parameters([in(_, _)|T]) :- !,
	write_out_parameters(T).
write_out_parameters([out(Type, Name)|T]) :-			% C-version
	setting(language, c), !,
	indirection(Type, client(return), Stars),
	atom_concat('_', Name, TmpVar),
	write_prolog_arg(in(TmpVar, Type, [pointer(Stars)]), Name),
	write_out_parameters(T).
write_out_parameters([out(Type, Name)|T]) :- !,			% C++-version
	atom_concat('_', Name, TmpVar),
	(   must_be_allocated(Type)
	->  Options = [var]
	;   Options = []
	),
	write_prolog_arg(in(TmpVar, Type, Options), Name),
	write_out_parameters(T).

write_rval_parameter(void) :- !.
write_rval_parameter(Type) :-					% C-version
	setting(language, c), !,
	indirection(Type, client(return), Stars),
	write_prolog_arg(in('_rval', Type, [pointer(Stars)]), rval).
write_rval_parameter(Type) :-					% C++-version
	(   must_be_allocated(Type)
	->  Options = [var]
	;   Options = []
	),
	write_prolog_arg(in('_rval', Type, Options), rval).

arity(Parms, void, Arity) :- !,
	length(Parms, Arity0),
	Arity is Arity0 + 1.
arity(Parms, _Return, Arity) :-
	length(Parms, Arity0),
	Arity is Arity0 + 2.

parameters([]).
parameters(Parms) :-
	utter([',', nl]),
	do_parms(Parms).
	
do_parms([H]) :- !,
	parameter(H).
do_parms([H|T]) :-
	parameter(H),
	utter([', ']),
	do_parms(T).

parameter(in(_, Name)) :-
	utter(['term_t ~w'+[Name]]).
parameter(out(_, Name)) :-
	utter(['term_t ~w'+[Name]]).
parameter(inout(_, Name)) :-
	utter(['term_t ~w'+[Name]]).

return_parameter(void).
return_parameter(_Type) :-
	utter([',', nl, 'term_t rval']).

	      
		 /*******************************
		 *	     EXCEPTIONS		*
		 *******************************/

client_exception_handler(Excepts) :-
	setting(language, cpp), !,
	corba_to_pl_exception,
	checklist(cpp_to_pl_exception, Excepts).
client_exception_handler(Excepts) :-
	utter(['static int', nl]),
	utter(['print_exception(CORBA_Environment *env)', nl]),
	utter([begin]),
	utter(['switch(env->_major)', nl]),
	utter([begin]),
	utter(['case CORBA_NO_EXCEPTION:', nl]),
	utter(['  return FALSE;', nl]),
	utter(['case CORBA_SYSTEM_EXCEPTION:', nl]),
	utter([begin]),
	utter(['char *id = CORBA_exception_id(env);', nl]),
	utter(['PL_CORBA_exception(PL_CORBA_SYSTEM_EXCEPTION, id);', nl]),
	utter(['CORBA_exception_free(env);', nl, nl]),
	utter(['return TRUE;', nl]),
	utter([end, nl]),
	(   Excepts \== []
	->  utter(['case CORBA_USER_EXCEPTION:', nl]),
	    utter([begin]),
	    utter(['char *id = CORBA_exception_id(env);', nl, nl]),
	    user_exceptions(Excepts),
	    utter(['CORBA_exception_free(env);', nl, nl]),
	    utter(['return TRUE;', nl]),
	    utter([end, nl])
	;   true
	),
	utter(['default:', nl]),
	utter(['  assert(0);', nl]),
	utter([end, nl]),
	utter([end, nl, nl]).

%	user_exceptions(+ExceptionList)
%
%	Creates te C-switch for exceptions for the C-binding.

user_exceptions([]) :- !,
	utter([begin]),
	utter(['PL_CORBA_exception(PL_CORBA_UNKNOWN_USER_EXCEPTION, id);',
	       nl]),
	utter([end, nl]).
user_exceptions([exception(Name, Scope, Members)|T]) :-
	concat_atom(Scope, '_', ScopeName),
	utter(['if ( env->returnCode == ex_~w_~w )'+[ScopeName, Name], nl]),
	utter([begin]),
	utter(['term_t et = PL_new_term_ref();', nl]),
	(   Members == []
	->  utter(['PL_unify_atom_chars(et, "~w");'+[Name], nl])
	;   utter(['term_t tmp = PL_new_term_ref();', nl]),
	    utter(['~w_~w *ctx = CORBA_exception_value(env);'+
		   [ScopeName, Name],
		   nl, nl]),
	    length(Members, Arity),
	    utter(['PL_unify_functor(et, PL_new_functor(PL_new_atom("~w"), ~d));'+
		   [Name, Arity], nl]),
	    write_struct_members(Members, in(ctx, _, [pointer(*)]), et, tmp)
	),
	utter(['PL_CORBA_exception(PL_CORBA_USER_EXCEPTION, et);', nl]),
	utter([end, ' else', nl]),
	user_exceptions(T).

corba_to_pl_exception :-
	utter(['static int', nl]),
	utter(['ex_CORBA_SystemException(CORBA(SystemException) &ex)', nl]),
	utter(begin),
	(   setting(ex_chars, CharsMember)
	->  utter(['const char *id = ex~w;'+[CharsMember], nl, nl])
	;   setting(orb, orbix)
	->  utter(['const char *id = "Orbix system exception";', nl, nl]),
	    utter(['cerr << "Orbix exception: " << ex;', nl])
	;   utter(['const char *id = ex;', nl, nl])
	),
	utter(['PL_CORBA_exception(PL_CORBA_SYSTEM_EXCEPTION, id);', nl, nl]),
	utter(['return FALSE;', nl]),
	utter([end, nl, nl]).


%	cpp_to_pl_exception(Exception)
%
%	Generates a C++ function to throw a Prolog exception from a CORBA
%	generated C++ exception.

cpp_to_pl_exception(exception(Name, Scope, Members)) :-
	exception_type(Name, Scope, TypeName),
	exception_function(Name, Scope, FunctionName),
	utter(['static int', nl]),
	utter(['~w(~w &ex)'+[FunctionName, TypeName], nl]),
	utter(begin),
	utter(['term_t et = PL_new_term_ref();', nl, nl]),
	utter(['term_t tmp = PL_new_term_ref();', nl]),
	length(Members, Arity),
	utter(['PL_unify_functor(et, PL_new_functor(PL_new_atom("~w"), ~d));'+
	       [Name, Arity], nl]),
	write_struct_members(Members, in(ex, exception, []), et, tmp),
	utter(nl),
	utter(['return PL_throw(et);', nl]),
	utter([end, nl, nl]).


%	exception_type(+ExceptionName, +Scope, -Name)
%	
%	Generate the C++ type-name of the named exception.

exception_type(Name, Scope, Type) :-
	setting(scope_delimiter, Delim),
	append(Scope, [Name], NameList),
	concat_atom(NameList, Delim, Type).
exception_function(Name, Scope, Function) :-
	append([ex|Scope], [Name], NameList),
	concat_atom(NameList, '_', Function).


		 /*******************************
		 *        INITIALISATION	*
		 *******************************/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Force working exit under Windows NT/95.  Don't know why it doesn't work
without though.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

emit_quit :-
	utter(['static void', nl]),
	utter(['quit(int status, void *closure)', nl]),
	utter([begin]),
	utter(['ExitProcess(0);', nl]),
	utter([end, nl, nl]).
module_initialiser(Module) :-
	(   current_prolog_flag(windows, true)
	->  emit_quit
	;   true
	),
	utter(['/* Initialization */', nl, nl]),
	(   setting(language, cpp)
	->  utter('extern "C" ')
	;   true
	),
	utter(['install_t', nl]),
	utter(['PL_~w__Initialize()'+[Module], nl]),
	utter(begin),
	(   setting(orb, ilu)
	->  utter(['~w__Initialize();'+[Module], nl, nl])
	;   true
	),
	(   current_prolog_flag(windows, true)
	->  utter(['PL_on_halt(quit, NULL);', nl, nl])
	;   true
	),
	register_foreign_predicates,
	utter([end, nl, nl]).
