/*  $Id: cutil.pl,v 1.2 2001/01/31 10:07:41 jan Exp $

    Part of Prolog <-> CORBA interface
    Designed and implemented by Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/

:- module(cstub_util,
	  [ utter/1,			% +Spec

	    prolog_read_value/3,	% +In, +Out, +Env
	    pl_get_function/2,		% +Type, -FunctionName
	    case_n/3,			% +Case, +Scope, -CCaseVal

	    write_prolog_arg/3,		% +Type, +C, +Prolog
	    write_prolog_arg/2,		% +C, +Prolog
	    write_struct_members/4,	% +Members, +C, +Prolog, +ArgVar

	    foreign/2,			% +Name/Arity, +Function
	    reset_foreigns/0,
	    register_foreign_predicates/0,

	    local_var/2,		% +Base, -Name
	    reset_local_vars/0,

	    translate_types/1,		% +Types

	    enum_def_name/3,		% +Enum, +Scope, +DefName
	    has_var_type/1,		% +Type
	    must_be_allocated/1,	% +Type
	    must_be_allocated/2,	% +Where, +Type
	    c_type/2,			% +IDL, -C
	    c_type/3,			% +IDL, +Where, -C
	    classify_type/2,		% +Type, -TypeClass
	    indirection/3,		% +Type, +Location, -Stars
	    named_type/3,		% +Name, Scope, -Identifier

	    include_files/1,		% +Module
	    file_base/2			% +Module, -Base
	  ]).
:- use_module(util).

:- op(50, xf, &).

prolog_read_value(In, Out, Env) :-
	generate_read_value(In, Out, Env).

%	out(VarName, Type, [Options])
%
%	This term contains info on the current (C) output.  Options:
%      
%	allocate	The output should be allocated first
%	pointer(Stars)	(C): argument indirection levels

out_name(out(VarName, _, _), VarName).
out_type(out(_, Type, _), Type).
out_option(out(_, _, Options), Option) :-
	memberchk(Option, Options).

generate_read_value(In, Out, Env) :-
	out_type(Out, type(Enum, Scope, enum(_))), !,
	out_ptr_name(Out, OutVar),
	enum_def_name(Enum, Scope, Def),
	utter(['PL_CORBA_get_enum(~w, (CORBA_enum*)~w, &~w, ~w);'+
	       [In, OutVar, Def, Env], nl]).
generate_read_value(In, Out, Env) :-	% STRUCT
	out_type(Out, Type),
	Type = type(_, _, struct(Members)), !,
	initialise_out(Out),
	local_var(arg, ArgVar),
	utter(begin),
	utter(['term_t ~w = PL_new_term_ref();'+[ArgVar], nl, nl]),
	prolog_read_struct(Members, 1, In, Out, Env, ArgVar),
	utter([end, nl]).
generate_read_value(In, Out, Env) :-	% UNION
	out_type(Out, Type),
	Type = type(Union, Scope, union(SType, Members)), !,
	initialise_out(Out),
	utter(begin),
	local_var(at, At),
	local_var(av, Av),
	utter(['term_t ~w = PL_new_term_ref();'+[At], nl]),
	utter(['term_t ~w = PL_new_term_ref();'+[Av], nl]),
	utter(['if ( PL_CORBA_get_union(~w, "~w", ~w, ~w, ~w) )'+
	       [In, Union, At, Av, Env], nl]),
	utter([begin]),
	c_type(SType, CSType),
	local_var(d, SwitchVar),
	utter(['~w ~w;'+[CSType, SwitchVar], nl, nl]),
	prolog_read_value(At, out(SwitchVar, SType, []), Env),
	(   setting(language, c)
	->  out_member_access(Out, '_d', OutTypeField),
	    utter(['~w = ~w;'+[OutTypeField, SwitchVar], nl])
	;   true
	),
	utter(['switch( ~w )'+[SwitchVar], nl]),
	utter(begin),
	generate_union_cases(Members, Scope, Av, Out, Env),
	utter([end, nl]),
	utter([end, nl]),
	utter([end, nl]).
generate_read_value(In, Out, Env) :-	% SEQUENCE
	out_type(Out, Type),
	Type = type(Seq, Scope, sequence(ItemType, _Size)), !,
	initialise_out(Out),
	utter(begin),
	utter(['term_t list = PL_copy_term_ref(~w);'+[In], nl]),
	utter(['term_t head = PL_new_term_ref();', nl]),
	(   setting(language, cpp)
	->  utter(['CORBA(ULong) len = 0;', nl]) % ORBIX
	;   true
	),
	utter(nl),
	utter(['while ( PL_get_list(list, head, list) )', nl]),
	utter(begin),
	c_type(ItemType, var, CItemType),
	local_var(item, ItemVar),
	utter(['~w ~w;'+[CItemType, ItemVar], nl, nl]),
	(   setting(language, c),
	    must_be_allocated(ItemType)
	->  Options = [allocate, pointer(*)]
	;   Options = []
	),
	prolog_read_value(head, out(ItemVar, ItemType, Options), Env),
	out_name(Out, Var),
	(   setting(language, cpp)
	->  (   out_option(Out, pointer(*)) 		% input arguments
	    ->	MemSep = '->',
		sformat(LenVar, '(*~w)', [Var])
	    ;	out_option(Out, var)			% local/output
	    ->	MemSep = '->',
		LenVar = Var
	    ;	MemSep = '.',
		LenVar = Var
	    ),
	    utter(['~w~wlength(len+1);'+[Var, MemSep], nl]),
	    utter(['~w[len++] = ~w;'+[LenVar, ItemVar], nl])
	;   concat_atom(Scope, '_', ScopeName),
	    out_ptr_name(Out, SeqVar),
	    utter(['~w_~w_Append(~w, ~w);'+
		   [ScopeName, Seq, SeqVar, ItemVar], nl])
	),
	utter([end, nl]),
	utter([end, nl]).
generate_read_value(In, Out, Env) :-
	out_type(Out, Type),
	pl_get_function(Type, PlFunc), !,
	out_ptr_name(Out, Var),
	utter(['~w(~w, ~w, ~w);'+[PlFunc, In, Var, Env], nl]).

prolog_read_struct([], _, _, _, _, _).
prolog_read_struct([Name:Type|T], Nth, In, Out, Env, ArgVar) :-
	utter(['PL_get_arg(~d, ~w, ~w);'+[Nth, In, ArgVar], nl]),
	NN is Nth + 1,
	out_member_access(Out, Name, Member),
	mkmember_out(Type, Member, MemberOut),
	prolog_read_value(ArgVar, MemberOut, Env),
	prolog_read_struct(T, NN, In, Out, Env, ArgVar).

generate_union_cases([], _, _, _, _).
generate_union_cases([case(N, Type, Name)|T], Scope, In, Out, Env) :-
	case_n(N, Scope, I),
	utter(['case ~w:'+[I], nl, begin]),
	(   setting(language, c)
	->  atom_concat('_u.', Name, Field),
	    out_member_access(Out, Field, Member),
	    mkmember_out(Type, Member, MemberOut),
	    prolog_read_value(In, MemberOut, Env)
	;   c_type(Type, var, TypeName),
	    utter(['~w _tmp;'+[TypeName], nl, nl]),
	    prolog_read_value(In, out('_tmp', Type, []), Env),
	    out_member_access(Out, Name, OutMem),
	    utter(['~w(_tmp);'+[OutMem, Name], nl])
	),
	utter(['break;', nl, end, nl]),
	generate_union_cases(T, Scope, In, Out, Env).

%	out_ptr_name(+Out, -Pointer)
%
%	Get a suitable output pointer for (primitive) types.  Prefix with
%	& if the argument is not a pointer

out_ptr_name(Out, VarName) :-
	out_option(Out, pointer(Stars)), !,
	out_name(Out, VarName0),
	mk_simple_pointer(Stars, VarName0, VarName).
out_ptr_name(Out, Ptr) :-
	out_name(Out, VarName),
	atom_concat(&, VarName, Ptr).

mk_simple_pointer('', Name0, Name) :- !,
	atom_concat(&, Name0, Name).
mk_simple_pointer(*,  Name,  Name).
mk_simple_pointer(**, Name0, Name) :-
	atom_concat(*, Name0, Name).

mkmember_out(Type, Out, out(Out, Type, [])) :-
	classify_type(Type, scalar), !.
mkmember_out(Type, Out, out(Out, Type, [])) :-
	classify_type(Type, string), !.
mkmember_out(Type, Out, out(Out, Type, [])) :-
	inlined_type(Type), !.
mkmember_out(Type, Out, out(Out, Type, [pointer(*)])).

out_member_access(Out, Field, Access) :-
	out_option(Out, pointer(*)), !,
	out_name(Out, Var),
	sformat(Access, '~w->~w', [Var, Field]).
out_member_access(Out, Field, Access) :-
	out_option(Out, var), !,
	out_name(Out, Var),
	sformat(Access, '~w->~w', [Var, Field]).
out_member_access(Out, Field, Access) :-
	out_option(Out, pointer(**)), !,
	out_name(Out, Var),
	sformat(Access, '(*~w)->~w', [Var, Field]).
out_member_access(Out, Field, Access) :-
	out_name(Out, Var),
	sformat(Access, '~w.~w', [Var, Field]).

%	initialise_out(+Out)
%
%	Prepare the output for writing to it.  Allocate if necessary, or
%	initialise the already allocated data.

initialise_out(Out) :-
	out_option(Out, allocate),
	setting(language, c), !,		% C-version
	out_type(Out, Type),
	out_ptr_name(Out, Var),
	Type = type(TypeName, Scope, _),
	(   classify_type(Type, sequence)
	->  Postfix = 'Create',
	    Args = '0, NULL'
	;   Postfix = '_alloc',
	    Args = ''
	),
	concat_atom(Scope, '_', ScopeName),
	utter([ '~w = ~w_~w_~w(~w);'+
		[Var, ScopeName, TypeName, Postfix, Args],
		nl
	      ]).
initialise_out(Out) :-
	out_option(Out, allocate), !,		% C++-version
	out_type(Out, Type),
	out_name(Out, Var),
	c_type(Type, Class),
	utter(['~w = new ~w;'+[Var, Class], nl]).
initialise_out(Out) :-
	setting(language, c),
	out_type(Out, Type),
	classify_type(Type, sequence), !,
	out_name(Out, Var),
	Type = type(TypeName, Scope, _),
	concat_atom(Scope, '_', ScopeName),
	utter(['~w_~w_Init(&~w, 0, NULL);'+[ScopeName, TypeName, Var], nl]).
initialise_out(_).


case_n(I, _, I) :-
	integer(I), !.
case_n(@(Name, _I), Scope, GlobalName) :- !,
	(   setting(scoped_enum, true)
	->  append(Scope, [Name], L),
	    concat_atom(L, '_', GlobalName)
	;   GlobalName = Name
	).
case_n(Case, _, _) :-
	error(illegal_union_case_value(Case)),
	fail.

pl_get_function(Type, PlFunc) :-
	do_pl_get_function(Type, PlFunc), !.
pl_get_function(Type, PlFunc) :-
	error(pl_get_function(Type, PlFunc)),
	fail.

do_pl_get_function(type(_,_,T), Func) :- !,
	do_pl_get_function(T, Func).
do_pl_get_function(scalar(unsigned(long)),   'PL_CORBA_get_ulong').
do_pl_get_function(scalar(unsigned(short)),  'PL_CORBA_get_ushort').
do_pl_get_function(scalar(long),   	     'PL_CORBA_get_long').
do_pl_get_function(scalar(short),  	     'PL_CORBA_get_short').
do_pl_get_function(scalar(octet),  	     'PL_CORBA_get_octet').
do_pl_get_function(scalar(double), 	     'PL_CORBA_get_double').
do_pl_get_function(scalar(float),  	     'PL_CORBA_get_float').
do_pl_get_function(string(_),		     'PL_CORBA_get_string').
do_pl_get_function(interface,		     'PL_CORBA_get_object').
do_pl_get_function(boolean,		     'PL_CORBA_get_boolean').

		 /*******************************
		 *	    C --> Prolog	*
		 *******************************/

write_prolog_arg(Type, Name, Arg) :-
	write_prolog_arg(in(Name, Type, []), Arg).

write_prolog_arg(In, Arg) :-
	do_write_prolog_arg(In, Arg), !.
write_prolog_arg(In, _) :-
	in_type(In, Type),
	error(write_argument(Type)),
	fail.

%	in(VarName, Type, [Options])
%
%	This term contains info on the current (C) input.  Options:
%      
%	allocate	The input should be allocated first
%	ptr		The input variable denotes a pointer

in_name(in(VarName, _, _), VarName).
in_type(in(_, Type, _), Type).
in_option(in(_, _, Options), Option) :-
	memberchk(Option, Options), !.
in_option(in(_, _, Options), pointer('')) :-
	\+ memberchk(Options, pointer(_)).

do_write_prolog_arg(In, Arg) :-
	in_type(In, type(Enum,Scope,enum(_))), !,
	in_value_name(In, InVar),
	enum_def_name(Enum, Scope, Def),
	utter(['PL_CORBA_unify_enum(~w, ~w, &~w);'+[Arg, InVar, Def], nl]).
do_write_prolog_arg(In, Arg) :-
/*	classify_type(In, sequence),
	(   setting(language, c)
	->  type_name(In, CTypeName)
	;   true
	), */
	in_type(In, type(Seq,Scope,sequence(ItemType, _Len))), !,
	concat_atom(Scope, '_', ScopeName),
	utter(begin),
	utter(['term_t head = PL_new_term_ref();', nl]),
	utter(['term_t tail = PL_copy_term_ref(~w);'+[Arg], nl]),
	(   setting(language, c)
	->  in_ptr_name(In, PtrName),
	    utter('int n, l = ~w_~w_Length(~w);'+[ScopeName, Seq, PtrName])
	;   member_access(In, 'length()', LenAccess),
	    utter('CORBA(ULong) n, l = ~w;'+[LenAccess])
	),
	utter([nl, nl]),
	utter(['for(n=0; n<l; n++)', nl]),
	utter(begin),
	local_var(elm, Elm),
	(   setting(language, c)
	->  c_type(ItemType, var, CItemType),
	    indirection(ItemType, var, Star0),
	    sequence_element_indirection(ItemType, Star),
	    utter('~w ~w~w = ~w_~w_Nth(~w, n);'+
		  [CItemType, Star, Elm, ScopeName, Seq, PtrName]),
	    atom_concat(Star0, Star, Stars),
	    Options = [pointer(Stars)]
	;   in_name(In, Var),
	    c_type(ItemType, part, CItemType),
	    utter('~w ~w = ~w[n];'+[CItemType, Elm, Var]),
	    Options = []
	),
	utter([nl, nl]),
	utter(['PL_unify_list(tail, head, tail);', nl]),
	write_prolog_arg(in(Elm, ItemType, Options), head),
	utter([end, nl]),
	utter(['PL_unify_nil(tail);', nl]),
	utter([end, nl]).
do_write_prolog_arg(In, Arg) :-
	in_type(In, type(Struct, _, struct(Members))), !,
	length(Members, Arity),
	utter(['PL_unify_functor(~w, PL_new_functor(PL_new_atom("~w"), ~d));'+
	       [Arg, Struct, Arity], nl]),
	utter(begin),
	local_var(arg, ArgVar),
	utter(['term_t ~w = PL_new_term_ref();'+[ArgVar], nl]),
	write_struct_members(Members, In, Arg, ArgVar),
	utter([end, nl]).
do_write_prolog_arg(In, Arg) :-
	in_type(In, type(Union, Scope, union(Switch, Cases))), !,
	utter(begin),
	local_var(at, At),
	local_var(av, Av),
	utter(['term_t ~w = PL_new_term_ref();'+[Av], nl]),
	utter(['term_t ~w = PL_new_term_ref();'+[At], nl]),
	(   setting(language, c)
	->  member_access(In, '_d', TypeArg)
	;   member_access(In, '_d()', TypeArg)
	),
	write_prolog_arg(in(TypeArg, Switch, []), At),
	utter([nl, 'switch( ~w )'+[TypeArg], nl]),
	utter(begin),
	write_union_cases(Cases, Scope, In, Av),
	utter([end, nl]),
	utter(['PL_CORBA_unify_union(~w, "~w", ~w, ~w);'+[Arg, Union, At, Av],
	       nl]),
	utter([end, nl]).
do_write_prolog_arg(in(Name, type(_,_,T), Options), Arg) :- !,
	do_write_prolog_arg(in(Name, T, Options), Arg).
do_write_prolog_arg(In, Arg) :-
	in_type(In, T),
	in_value_name(In, Name),
	write_simple_prolog_arg(T, Name, Arg).

write_simple_prolog_arg(scalar(float), Name, Arg) :- !,
	utter(['PL_unify_float(~w, ~w);'+[Arg, Name], nl]).
write_simple_prolog_arg(scalar(double), Name, Arg) :- !,
	utter(['PL_unify_float(~w, ~w);'+[Arg, Name], nl]).
write_simple_prolog_arg(scalar(_), Name, Arg) :-
	utter(['PL_unify_integer(~w, ~w);'+[Arg, Name], nl]).
write_simple_prolog_arg(string(_), Name, Arg) :-
	utter(['PL_unify_atom_chars(~w, ~w);'+[Arg, Name], nl]).
write_simple_prolog_arg(boolean, Name, Arg) :-
	utter(['PL_CORBA_unify_boolean(~w, ~w);'+[Arg, Name], nl]).
write_simple_prolog_arg(interface, Name, Arg) :-
	utter(['PL_CORBA_unify_interface(~w, ~w);'+[Arg, Name], nl]).

in_ptr_name(Out, VarName) :-		% TBD: Unify with out_ptr_name!
	in_option(Out, pointer(Stars)), !,
	in_name(Out, VarName0),
	mk_simple_pointer(Stars, VarName0, VarName).
in_ptr_name(Out, Ptr) :-
	in_name(Out, VarName),
	atom_concat(&, VarName, Ptr).

in_value_name(Out, VarName) :-
	in_option(Out, pointer(Stars)), !,
	in_name(Out, VarName0),
	atom_concat(Stars, VarName0, VarName).
in_value_name(Out, VarName) :-
	in_name(Out, VarName).

write_union_cases([], _, _, _).
write_union_cases([case(N, Type, Name)|T], Scope, In, Out) :-
	case_n(N, Scope, I),
	utter(['case ~w:'+[I], nl]),
	utter(indent(2)),
	(   setting(language, c)
	->  member_access(In, '_u', InUMem),
	    concat_atom([InUMem, Name], '.', InMem),
	    mkmember_in(Type, InMem, MemberIn)
	;   atom_concat(Name, '()', Field),
	    member_access(In, Field, InMem),
	    mkmember_in(Type, InMem, MemberIn)
	),
	write_prolog_arg(MemberIn, Out),
	utter(['break;', nl, undent, nl]),
	write_union_cases(T, Scope, In, Out).

%	struct in struct is not a pointer.  To enforce a pointer, we
%	use &.  This is not based on calssify type, but whether this
%	this is correct?

write_struct_members(Members, C, Pl, ArgVar) :-
	write_struct_members(Members, 1, C, Pl, ArgVar).

write_struct_members([], _, _, _, _).
write_struct_members([Name:Type|T], Nth, In, Pl, ArgVar) :-
	utter(['PL_get_arg(~d, ~w, ~w);'+[Nth, Pl, ArgVar], nl]),
	member_access(In, Name, CMem),
	mkmember_in(Type, CMem, MemberIn),
	write_prolog_arg(MemberIn, ArgVar),
	NN is Nth+1,
	write_struct_members(T, NN, In, Pl, ArgVar).

member_access(In, Name, Mem) :-
	in_option(In, var), !,
	in_name(In, InVar),
	sformat(Mem, '~w->~w', [InVar, Name]).
member_access(In, Name, Mem) :-
	setting(language, cpp), !,
	in_name(In, InVar),
	sformat(Mem, '~w.~w', [InVar, Name]).
member_access(In, Field, Mem) :-
	in_option(In, pointer(*)), !,
	in_name(In, Var),
	sformat(Mem, '~w->~w', [Var, Field]).
member_access(In, Field, Mem) :-
	in_option(In, pointer('')), !,
	in_name(In, Var),
	sformat(Mem, '~w.~w', [Var, Field]).

mkmember_in(Type, In, in(In, Type, [])) :-
	classify_type(Type, scalar), !.
mkmember_in(Type, In, in(In, Type, [])) :-
	classify_type(Type, string), !.
mkmember_in(Type, In, in(In, Type, [])) :-
	inlined_type(Type), !.
mkmember_in(Type, In, in(In, Type, [pointer(*)])).

%	type_name(Type, CTypeName)
%
%	Determine the name of the (C-)type from an IDL type.

type_name(type(Alias, Scope, _), Name) :- !,
	append(Scope, [Alias], NameParts),
	concat_atom(NameParts, '_', Name).
type_name(type(sequence(ItemType, _Len)), Name) :- !,
	type_name(ItemType, ItemTypeName),
	concat_atom(['CORBA_sequence', ItemTypeName], '_', Name).


		 /*******************************
		 *	       TYPES		*
		 *******************************/

translate_types(Types) :-
	forall(member(Type, Types),
	       translate_type(Type)).

translate_type(type(Enum, Scope, enum(Members))) :- !,
	enum_def_name(Enum, Scope, Def),
	sformat(CharArrayName, '~w_names', Def),
	utter(['static const char *~w[] ='+[CharArrayName], nl]),
	utter(begin),
	emit_enum_members(Members),
	utter([end, ';', nl]),
	utter(['static PL_CORBA_enum_def ~w =~n  { "~w", ~w };'+
	       [Def, Enum, CharArrayName], nl, nl]).
translate_type(_).	

emit_enum_members([]) :-
	utter(['(char *)0', nl]).
emit_enum_members([H|T]) :-
	utter(['"~w",'+[H], nl]),
	emit_enum_members(T).


		 /*******************************
		 *	    TYPE LOGIC		*
		 *******************************/

%	has_var_type(+Type)
%
%	True if, for the C++ mapping, the type has an associated <T>_var
%	type.

has_var_type(Type) :-
	setting(language, cpp),
	classify_type(Type, Class),
	var_class(Class).

var_class(string).
var_class(interface).
var_class(record(_)).
var_class(union(_)).
var_class(sequence).
var_class(array(_)).

%	must_be_allocated(Type)
%
%	Succeeds if a type needs to be allocated for output and return
%	arguments.

must_be_allocated(Type) :-
	classify_type(Type, Class),
	var_class(Class).

must_be_allocated(out, Type) :-
	must_be_allocated(Type),
	classify_type(Type, Class),
	\+ memberchk(Class, [record(fixed), union(fixed)]).

sequence_element_indirection(Type, '') :-
	classify_type(Type, Classification),
	pointer_in_sequence(Classification), !.
sequence_element_indirection(_, *).

pointer_in_sequence(record(var)).
pointer_in_sequence(union(var)).

enum_def_name(Enum, Scope, DefName) :-
	concat_atom(Scope, '_', ScopeName),
	concat_atom([enum, ScopeName, Enum], '_', DefName).

inlined_type(type(_, _, struct(_))).
inlined_type(type(_, _, union(_, _))).
inlined_type(type(_, _, sequence(_, _))).

c_type(void, return, void) :- !.
c_type(Type, Where, CType) :-
	setting(language, c), !,
	c_type(Type, CType0),
	indirection(Type, Where, Stars),
	atom_concat(CType0, Stars, CType).
c_type(Type, return, CType) :-
	setting(language, cpp),
	classify_type(Type, interface), !,
	c_type(Type, CType0),
	atom_concat(CType0, '_ptr', CType).
c_type(Type, return(var), 'CORBA(Object_ptr)') :-
	setting(language, cpp),
	classify_type(Type, interface), !.
c_type(Type, return(var), CType) :-
	setting(has_retn, true),
	has_var_type(Type), !,
	c_type(Type, CType0),
	mkvar_type(CType0, CType).
c_type(Type, return(var), CType) :- !,
	c_type(Type, return, CType).
c_type(Type, client(return), CType) :-
	setting(language, cpp),
	classify_type(Type, interface), !,
	c_type(Type, CType0),
	atom_concat(CType0, '_ptr', CType).
c_type(Type, client(_), CType) :-
	has_var_type(Type), !,
	c_type(Type, CType0),
	mkvar_type(CType0, CType).
c_type(Type, client(_), CType) :- !,
	c_type(Type, return, CType).
c_type(Type, out, CType) :-
	setting(has_out, true), !,
	(   classify_type(Type, string)
	->  CType = 'CORBA(String_out)'
	;   c_type(Type, CType0),
	    atom_concat(CType0, '_out', CType)
	).
c_type(Type, Where, CType) :-		% probably incomplete
	c_type(Type, CType0),
	classify_type(Type, Kind),
	cppdecl(Where, Kind, Decls),
	make_var_type(Type, CType0, CType1),
	declcpp(Decls, CType1, CType).

declcpp('', T, T).
declcpp(const&, T0, T) :-
	sformat(T, 'const ~w&', [T0]).
declcpp(const, T0, T) :-
	sformat(T, 'const ~w', [T0]).
declcpp(*, T0, T) :-
	sformat(T, '~w*', [T0]).
declcpp(&, T0, T) :-
	sformat(T, '~w&', [T0]).
declcpp(*&, T0, T) :-
	sformat(T, '~w*&', [T0]).

make_var_type(Type, T0, T) :-
	classify_type(Type, interface), !,
	mkvar_type(T0, T).
make_var_type(_Type, T, T).

mkvar_type('CORBA_String', 'CORBA(String_var)') :- !.
mkvar_type(T0, T) :-
	concat(T0, '_var', T).


%	ctype(+Type, -CType)
%
%	Determine the basic type identifier in the C/C++ mapping for an
%	IDL type.  Note that strings are always mapped on CORBA_String.
%	This is at least necessary for Omnibroker/g++.  It sounds more
%	logical to use the name-alias in case of a typedef string foo;
%	To select this, transpose the first two clauses.

c_type(StringAlias, 'CORBA_String') :-
	classify_type(StringAlias, string), !.
c_type(type(Name, Scope, _), CType) :- !,
	named_type(Name, Scope, CType).
c_type(Type, CType) :-
	primitive_type(Type, C, CPP), !,
	(   setting(language, c)
	->  T0 = C
	;   T0 = CPP
	),
	setting(scope_delimiter, Delim),
	concat_atom(['CORBA', T0], Delim, CType).
c_type(Type, _) :-
	error(no_c_type(Type)),
	fail.

primitive_type(scalar(long),		long,		'Long').
primitive_type(scalar(short),		short,		'Short').
primitive_type(scalar(unsigned(long)),  unsigned_long,	'ULong').
primitive_type(scalar(unsigned(short)),	unsigned_short,	'UShort').
primitive_type(scalar(float),		float,		'Float').
primitive_type(scalar(double),		double,		'Double').
primitive_type(scalar(char),		char,		'Char').
primitive_type(scalar(octet),		octet,		'Octet').
primitive_type(boolean,			boolean,	'Boolean').

named_type(Name, Scope, CType) :-
	setting(namespaces, false), !,
	append(Scope, [Name], Parts),
	setting(scope_delimiter, Delim),
	concat_atom(Parts, Delim, CType).
named_type(Name, _Scope, Name) :-
	setting(namespaces, true), !.
named_type(Name, Scope, CType) :-
	append(Scope, [Name], Parts),
	setting(scope_delimiter, Delim),
	concat_atom(Parts, Delim, CType).

classify_type(type(_,_,T), ParamType) :- !,
	classify_type(T, ParamType).
classify_type(Type, ParamType) :-
	parameter_type(Type, ParamType0), !,
	ParamType = ParamType0.

indirection(Type, client(Location), Stars) :- !,
	indirection(Type, Location, Stars).
indirection(Type, Location, Stars) :-
	classify_type(Type, ParamType), 
	stars(Location, ParamType, Stars), !.

parameter_type(scalar(_),	scalar).
parameter_type(char,		scalar).
parameter_type(octet,		scalar).
parameter_type(boolean,		scalar).
parameter_type(enum(_),		scalar).
parameter_type(string(_),	string).
parameter_type(sequence(_, _),	sequence).
parameter_type(struct(Members),	record(VarLength)) :-
	variable_length(Members, VarLength).
parameter_type(union(_, Members), union(VarLength)) :-
	variable_length(Members, VarLength).
parameter_type(interface,	interface).
parameter_type('Object',	interface).
parameter_type(T,		_) :-
	format(user_error, 'Don''t know how to classify type "~w"~n', [T]),
	fail.

variable_length(Members, var) :-
	member(_:Type, Members),
	variable_length_type(Type), !.
variable_length(Members, var) :-
	member(case(_N, Type, _Name), Members),
	variable_length_type(Type), !.
variable_length(_, fixed).

variable_length_type(Type) :-
	classify_type(Type, ParamType),
	variable_length_classification(ParamType).

variable_length_classification(record(var)).
variable_length_classification(union(var)).
variable_length_classification(string).
variable_length_classification(sequence).
variable_length_classification(array(var)).
variable_length_classification(interface). % ???
variable_length_classification(any).

stars(var,       Type, S) :- stars(Type, S, _, _, _, _).
stars(in,        Type, S) :- stars(Type, S, _, _, _, _).
stars(inout,     Type, S) :- stars(Type, _, S, _, _, _).
stars(out,       Type, S) :- stars(Type, _, _, S, _, _).
stars(return,    Type, S) :- stars(Type, _, _, _, S, _). % ???
stars(return(_), Type, S) :- stars(Type, _, _, _, S, _).
stars(exn,       Type, S) :- stars(Type, _, _, _, _, S).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%			In	InOut	Out	  Return    ExtN
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
stars(scalar,		'',	*,	*,	  '',	    *).
stars(record(fixed),	*,	*,	*,	  '',	    *).
stars(record(var),	*,	*,	**,	  *,	    *).
stars(union(fixed),	*,	*,	*,	  '',	    *).
stars(union(var),	*,	*,	**,	  *,	    *).
stars(string,		'',	*,	*,	  '',	    *).
stars(sequence,		*,	*,	**,	  *,	    *).
stars(interface,	'',	*,	*,	  '',	    *).
stars(array(fixed),	'',	'',	'',	  slice(*), *).
stars(array(var),	'',	'',	slice(*), slice(*), *).


cppdecl(var,    Type, S) :- cppdecl(Type, S, _, _, _, _).
cppdecl(in,     Type, S) :- cppdecl(Type, _, S, _, _, _).
cppdecl(inout,  Type, S) :- cppdecl(Type, _, _, S, _, _).
cppdecl(out,    Type, S) :- cppdecl(Type, _, _, _, S, _).
cppdecl(return, Type, S) :- cppdecl(Type, _, _, _, _, S).
cppdecl(part,	Type, S) :- cppdecl(Type, _, S, _, _, _).

cppdecl(scalar,		'', '',		&,	&,	  '').
cppdecl(record(fixed),	'', const&,	&,	&,	  '').
cppdecl(record(var),	'', const&,	&,	*&,	  *).
cppdecl(union(fixed),	'', const&,	&,	&,	  '').
cppdecl(union(var),	'', const&,	&,	*&,	  *).
cppdecl(string,		'', const,	&,	&,	  '').
cppdecl(sequence,	'', const&,	&,	*&,	  *).
cppdecl(interface,	'', '',		&,	&,	  '').
cppdecl(array(fixed),	'', const,	'',	'',	  slice(*)).
cppdecl(array(var),	'', const,	'',	slice(*)&,slice(*)).
cppdecl(any,		'', const&,	&,	*&,       *).


		 /*******************************
		 *	      FOREIGN		*
		 *******************************/

:- dynamic
	declared_foreign/2.		% Name, Arity

foreign(NameAndArity, Function) :-
	assert(declared_foreign(NameAndArity, Function)).

reset_foreigns :-
	retractall(declared_foreign(_, _)).

register_foreign_predicates :-
	forall(declared_foreign(Name/Arity, Function),
	       utter(['PL_register_foreign("~w", ~d, ~w, 0);'
		      +[Name, Arity, Function], nl])).


		 /*******************************
		 *	     LOCAL VARS		*
		 *******************************/

:- dynamic
	local_var_store/2.

reset_local_vars :-
	retractall(local_var_store(_, _)).

local_var(Name, Var) :-
	retract(local_var_store(Name, N)), !,
	NN is N + 1,
	concat(Name, NN, Var),
	asserta(local_var_store(Name, NN)).
local_var(Name, Name) :-
	asserta(local_var_store(Name, 1)).


		 /*******************************
		 *	      OUTPUT		*
		 *******************************/

utter([]) :- !.
utter([H|T]) :- !,
	utter(H),
	utter(T).
utter(nl) :- !,
	nl(cout).
utter(indent) :- !,
	setting(indent, X),
	line_position(cout, OldPos),
	Tabs is (X-OldPos) // 8,
	Spaces is (X-OldPos) mod 8,
	forall(between(1, Tabs, _), put(cout, 9)),
	forall(between(1, Spaces, _), put(cout, 32)).
utter(push(indent)) :- !,
	line_position(cout, Indent),
	push(indent, Indent).
utter(pop(indent)) :- !,
	pop(indent).
utter(begin) :- !,
	utter(indent),
	format(cout, '{ ', []),
	setting(indent, X),
	Y is X + 2,
	push(indent, Y).
utter(end) :- !,
	pop(indent),
	utter(indent),
	format(cout, '}', []).
utter(indent(Inc)) :- !,
	setting(indent, X),
	Y is X + Inc,
	push(indent, Y),
	utter(indent).
utter(undent) :- !,
	pop(indent),
	utter(indent).
utter(Fmt+Args) :- !,
	utter(indent),
	format(cout, Fmt, Args).
utter(Fmt) :-
	utter(indent),
	format(cout, Fmt, []).


		 /*******************************
		 *	       FILES		*
		 *******************************/

include_files(Module) :-
	setting(orb, ilu), !,
	utter(['#include <~w.h>'+[Module],nl]),
	utter(['#include <pl2ilu.h>', nl, nl]).
include_files(Module) :-
	setting(orb, omnibroker), !,
	file_base(Module, Base),
	utter(['#include <OB/CORBA.h>', nl]),
	utter(['#include <~w_skel.h>'+[Base],nl]),
	utter(['#include <pl2ilu.h>', nl, nl]).
include_files(Module) :-
	setting(orb, orbacus), !,
	file_base(Module, Base),
	utter(['#include <OB/CORBA.h>', nl]),
	utter(['#include <~w_skel.h>'+[Base],nl]),
	utter(['#include <pl2ilu.h>', nl, nl]).
include_files(Module) :-
	setting(orb, orbix), !,
	file_base(Module, Base),
	utter(['#include <CORBA.h>', nl]),
	utter(['#include <~w.h>'+[Base],nl]),
	utter(['#include <pl2ilu.h>', nl]),
	utter(['#include <ostream.h>', nl, nl]).
include_files(_Module) :-
	setting(orb, Orb),
	error(no_include_files_for_orb(Orb)).

file_base(_, Base) :-
	setting(file_base, input), !,
	setting(idl_file, IdlFile),
	file_name_extension(Base0, idl, IdlFile),
	localize_file(Base0, Base).
file_base(Base, Base).

localize_file(File0, File) :-
	absolute_file_name('', Pwd),
	concat(Pwd, File, File0), !.
localize_file(File, File).





