/*  $Id: cwrapper.pl,v 1.2 2001/01/25 15:28:36 jan Exp $

    Designed and implemented by Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/

:- module(corba_cstub,
	  [ cwrapper/1,
	    cppwrapper/1,
	    plstubber/0
	  ]).
:- use_module(idl).
:- use_module(plskel).
:- use_module(util).
:- use_module(cutil).
:- use_module(cclient).

default_orb(orbix22).			% default for now

plstubber :-
	reset_settings,
	current_prolog_flag(argv, Argv),
	append(_, ['--'|Args], Argv), !,
	default_orb(ORB),
	set_orb(ORB),
	(   handle_options(Args, RestArgs),
	    forall(member(Idl, RestArgs),
		   wrapper(Idl))
	->  halt(0)
	;   format(user_error, '~NERROR: Translation failed!~n', []),
	    halt(1)
	).

option('--orb', [orb=ORB], set_orb(ORB),
       'Select ORB (ilu|omnibroker|orbacus|orbix)').
option('-v',   [],    push(verbose, true),
       'Verbose operation').
option('--client', [], push(generate_client, true),
       'Generate <base>-clientstub.<ext>').
option('--help', [], help,
       'Provide help').
option('--fb', [input=Input], push(file_base(Input)),
       'Select output filebase (input,module)').
option('--debug', [subjects=Subjects], set_debug(Subjects),
       'Set debug options').
option('--spy', [predicates=Subjects], set_spy(Subjects),
       'Set Prolog spy-points (debugging)').

help :-
	format(user_error, 'Usage: pl-stubber [options] idl-file ...~n', []),
	forall(option(Name, Args, _, Comment),
	       help_on_option(Name, Args, Comment)).

help_on_option(Name, [], Comment) :- !,
	format(user_error, '~t~8|~w~t~32|~w~n',
	       [Name, Comment]).
help_on_option(Name, [OptName=_Var], Comment) :-
	format(user_error, '~t~8|~w=~w~t~32|~w~n',
	       [Name, OptName, Comment]).

option_name(Name=_Var, Name).

option_arguments([], L, L).
option_arguments([_Name=Var|TA], [Var|T0], T) :-
	option_arguments(TA, T0, T).

handle_options([], []).
handle_options(List, List) :-
	List = [Opt|_],
	\+ concat(-, _, Opt), !.
handle_options([LongOpt|L0], L) :-
	long_option(LongOpt), !,
	handle_options(L0, L).
handle_options([Opt|L0], L) :-
	option(Opt, Args, Goal, _Comment), !,
	option_arguments(Args, L0, L1),
	Goal,
	handle_options(L1, L).
handle_options([Opt0|T0], T) :-
	format(user_error, 'Unknown option: ~w', [Opt0]),
	handle_options(T0, T).

long_option(OptSpec) :-
	atom_codes(OptSpec, OptChars),
	phrase(long_option(Opt, Arg), OptChars),
	option(Opt, [_=Arg], Cmd, _Comment),
	Cmd.

long_option(Opt, Arg) -->
	[0'-, 0'-|OptChars],
	"=",
	[A0|A],
	end_of_string,
	{ atom_codes(Opt, [0'-, 0'-|OptChars]),
	  name(Arg, [A0|A])
	}.

end_of_string([], []).

set_debug(Options) :-
	term_to_atom(List, Options),
	set_debug_(List).

set_debug_((A,B)) :- !,
	set_debug_(A),
	set_debug_(B).
set_debug_(A) :-
	push(debugging, A).

set_spy(Options) :-
	term_to_atom(List, Options),
	set_spy_(List).

set_spy_((A,B)) :- !,
	set_spy_(A),
	set_spy_(B).
set_spy_(A) :-
	user:spy(A).

set_orb(Orb) :-
	setting(orb, Orb), !.
set_orb(omnibroker) :- !,
	push(language, cpp),
	push(impl_suffix, '-impl.cpp'),
	push(client_suffix, '-clientstub.cpp'),
	push(impl_class_suffix, '_impl'),
	push(skel_class_suffix, '_skel'),
	push(skel_class_prefix, ''),
	push(member_function_decl, ''),
	push(ex_chars, '._type_id()'),
	push(orb, omnibroker),
	push(namespaces, false),
	push(scoped_enum, true),
	push(scope_delimiter, '_'),
	push(has_retn, true),
	push(file_base, input).
set_orb(orbacus) :- !,
	push(language, cpp),
	push(impl_suffix, '-impl.cpp'),
	push(client_suffix, '-clientstub.cpp'),
	push(impl_class_suffix, '_impl'),
	push(skel_class_suffix, ''),
	push(skel_class_prefix, 'POA_'),
	push(member_function_decl, 'virtual '),
	push(ex_chars, '._to_string()'),
	push(orb, orbacus),
	push(shared_object, true),	% generate install().  Option?
	push(namespaces, true),
	push(have_bootmanager, true),
	push(scoped_enum, false),
	push(scope_delimiter, '::'),
	push(has_retn, true),		% _var types have ._retn()
	push(has_out, true),		% Output using <Type>_out
	push(file_base, input).
set_orb(ilu) :- !,
	push(language, c),
	push(impl_suffix, '-impl.c'),
	push(client_suffix, '-clientstub.c'),
	push(namespaces, false),
	push(scoped_enum, true),
	push(scope_delimiter, '_'),
	push(has_retn, false),
	push(orb, ilu),
	push(file_base, module).
set_orb(orbix) :- !,
	push(language, cpp),
	push(impl_suffix, '-impl.cpp'),
	push(client_suffix, '-clientstub.cpp'),
	push(impl_class_suffix, '_impl'),
	push(skel_class_suffix, 'BOAImpl'),
	push(skel_class_prefix, ''),
	push(member_function_decl, 'virtual '),
	push(orb, orbix),
	push(namespaces, false),
	push(scoped_enum, false),
	push(scope_delimiter, '_'),
	push(has_retn, false),
	push(file_base, input).
set_orb(orbix22) :- !,
	push(language, cpp),
	push(impl_suffix, '-impl.cpp'),
	push(client_suffix, '-clientstub.cpp'),
	push(impl_class_suffix, '_impl'),
	push(skel_class_suffix, 'BOAImpl'),
	push(skel_class_prefix, ''),
	push(member_function_decl, 'virtual '),
	push(orb, orbix),
	push(namespaces, true),
	push(scoped_enum, false),
	push(scope_delimiter, '::'),
	push(has_retn, false),
	push(file_base, input).
set_orb(ORB) :-
	error(no_defined_orb(ORB)),
	fail.


wrapper(Idl) :-
	find_input(Idl, InFile),
	push(idl_file, InFile),
	idl_file_to_term(InFile,
			 definition(Dcls,
				    Interfaces,
				    Modules)),
	reset_foreigns,
	foreign(corba_orb/1, pl_corba_orb),
	foreign(corba_initialize_server/2, pl_corba_initialize_server),
	foreign(corba_main_loop/1, pl_corba_main_loop),
	foreign(corba_object_to_string/2, pl_corba_object_to_string),
	foreign(corba_duplicate/2, pl_corba_duplicate),
	(   empty_dcls(Dcls),
	    Interfaces == []
	->  forall(member(M, Modules), progress(cwrapper_for_module(M)))
	;   file_name_extension(Base0, _, InFile),
	    file_base_name(Base0, Base),
	    format(user_error,
		   'Declarations outside modules, wrapping in top-module ~w~n',
		   [Base]),
	    Defs = definition(Dcls, Interfaces, Modules),
	    cwrapper_for_module(module(Base, Defs))
	),
	pop(idl_file).

find_input(Idl, InFile) :-
	absolute_file_name(Idl, [extensions(idl), access(read)], InFile), !.
find_input(Idl, _) :-
	error(no_file(Idl)),
	fail.

cwrapper(Idl) :-
	reset_settings,
	set_orb(ilu),
	push(generate_client, true),
	wrapper(Idl).

cppwrapper(Idl) :-
	reset_settings,
	set_orb(omnibroker),
	push(generate_client, true),
	wrapper(Idl).


empty_dcls(dcls([], [], [])).

cwrapper_for_module(module(Name, Defs)) :- % TBD: Push/Pop output streams
	(   setting(generate_client, true)
	->  progress(cclient_for_module(module(Name, Defs)))
	;   setting(impl_suffix, Suffix),		% make server
	    (   setting(file_base, input)		
	    ->  setting(idl_file, IdlFile),
		file_name_extension(Base, idl, IdlFile)
	    ;   Base = Name
	    ),
	    atom_concat(Base, Suffix, Impl),
	    atom_concat(Base, '-skel.pl', PlImpl),
	    open(Impl, write, CFd, [alias(cout)]),
	    open(PlImpl, write, PlFd, [alias(plout)]),
	    (   translate_module(module(Name, Defs))
	    ->  close(CFd),
		close(PlFd)
	    ;   close(CFd),
		close(PlFd),
		fail
	    )
	).

translate_module(module(Name,
			definition(dcls(_Consts, % TBD
					Types,
					Except),
				   Interfaces,
				   _Modules))) :- % TBD
	push(module, Name),
	pl_header(Name),
	utter('/* Module ~w */~n~n'+[Name]),
	include_files(Name),
	(   setting(namespaces, true)
	->  utter(['namespace ~w'+[Name], nl, '{', nl])
	;   true
	),
	progress(translate_types(Types)),
	progress(translate_exceptions(Except)),
	forall(member(Itf, Interfaces), progress(translate_interface(Itf))),
	progress(module_initialiser(Name)), % scope?
	(   setting(namespaces, true)
	->  utter([nl, '}', nl])
	;   true
	),
	pop(module).

translate_interface(interface(Name,
			      Scope,
			      _dcls,	% TBD
			      _Attrs,	% TBD
			      Methods)) :-
	utter('/* Interface ~w */~n~n'+[Name]),
	pl_interface_header(Name),
	cpp_class(Name, Scope, Methods),
	create_predicate(Name, Scope),
	utter('/* Methods for Interface ~w */~n~n'+[Name]),
	forall(member(M, Methods),
	       (   true, /*pl_skeleton(M),*/
		   progress(function_wrapper(Name, Scope, M)),
		   pl_method_skeleton(Name, Scope, M)
	       )).

create_predicate(Itf, Scope) :-
	flatten([pl, corba, Scope, Itf, create, true], Parts),
	concat_atom(Parts, '_', CFunc),
	concat_atom([Itf, '_create_true'], Predicate),
	utter(['static foreign_t', nl]),
	utter(['~w(term_t name, term_t self)'+[CFunc], nl]),
	utter(begin),
	utter(['char *id = NULL;', nl]),
	create_itf(Itf, Scope),
	utter(['return PL_unify_pointer(self, obj);', nl]),
	utter([end, nl, nl]),
	foreign(Predicate/2, CFunc).

cpp_class(Itf, Scope, Methods) :-
	setting(language, cpp), !,
	cpp_type_and_class(Itf, Scope, _Type, Class),
	setting(impl_class_suffix, Suffix),
	setting(skel_class_suffix, Skel),
	setting(skel_class_prefix, SkelPref),
	utter(['class ~w~w'+[Class,Suffix]]),
	(   setting(orb, omnibroker)
	->  utter([' : virtual public ~w~w~w'+[SkelPref, Class, Skel], nl])
	;   setting(orb, orbacus)
	->  setting(module, Module),
	    utter([' : virtual public POA_~w::~w'+[Module, Class], nl])
	;   utter([' : public ~w~w~w'+[SkelPref, Class, Skel], nl])
	),
	utter(begin),
	utter(['public:', nl]),
	(   setting(orb, orbacus)
	->  true
	;   utter(['~w~w() {}'+[Class, Suffix], nl]),
	    utter(['~w~w(const char *id) : ~w~w(id) {}'+
		   [Class, Suffix, Class, Skel],
		   nl, nl])
	),
	forall(member(M, Methods), cpp_prototype(M)),
	utter([end, ';', nl, nl]).
cpp_class(_, _, _).


cpp_prototype(method(Name, Parms, Return, _Raises, _Context)) :-
	c_type(Return, return, CReturn),
	setting(member_function_decl, Decl),
	utter(['~w~w ~w('+[Decl, CReturn, Name]]),
	utter(push(indent)),
	maplist(c_param, Parms, CParms),
	write_c_function_args(CParms),
	write_env_arg(Parms),
	utter([');', nl]),
	utter(pop(indent)).


create_itf(Itf, Scope) :-
	setting(language, c), !,
	flatten([Scope, Itf, '_CreateTrue'], Parts),
	concat_atom(Parts, '_', CreateFunc),
	utter(['void *obj;', nl, nl]),
	utter(['PL_get_atom_chars(name, &id);', nl]),
	utter(['obj = ~w(id, PL_CORBA_default_server(), NULL);'+[CreateFunc],
	       nl, nl]).
create_itf(Itf, Scope) :-
	setting(orb, orbacus), !,
	cpp_type_and_class(Itf, Scope, _Type, Class),
	setting(impl_class_suffix, Suffix),
	utter(['~w~w *impl;'+[Class, Suffix], nl, nl]),
	utter(['impl = new ~w~w();'+[Class, Suffix], nl]),
	utter(['CORBA(Object_ptr) obj = impl->_this();', nl]),
	utter(['PL_CORBA_bind(obj, name);', nl, nl]).
create_itf(Itf, Scope) :-
	cpp_type_and_class(Itf, Scope, _Type, Class),
	utter(['CORBA(Object_ptr) obj;', nl, nl]), % ORBIX
	utter(['if ( PL_get_atom_chars(name, &id) )', nl, indent(2)]),
	setting(impl_class_suffix, Suffix),
	utter(['obj = new ~w~w(id);'+[Class, Suffix], undent, nl]),
	utter([else, nl, indent(2)]),
	utter(['obj = new ~w~w();'+[Class, Suffix], undent, nl, nl]).
	      
cpp_type_and_class(Itf, Scope, Type, Class) :-
	c_type(type(Itf, Scope, interface), var, Type),
	(   setting(namespaces, true)
	->  Class = Itf
	;   setting(scope_delimiter, Delim),
	    append(Scope, [Itf], Parts),
	    concat_atom(Parts, Delim, Class)
	).

function_wrapper(Itf, Scope, method(Name, Parms, Return, Raises, _Context)) :-
	reset_local_vars,
	impl_function(Scope, Itf, Name, FunctionName),
	concat_atom(Scope, '_', ScopeName),
	c_type(Return, return, CReturn),
	maplist(c_param, Parms, CParms),
					% write the C/C++ function header
	utter('~w~n'+[CReturn]),
	utter('~w('+[FunctionName]),
	utter(push(indent)),
	write_this_arg(Itf, Scope),
	write_c_function_args(CParms),
	write_env_arg(Parms),
	utter([')', nl]),
	utter(pop(indent)), 

					% setup Prolog call
	length(Parms, PlArgc0),
	(   Return == void
	->  PlArgc is PlArgc0 + 1
	;   PlArgc is PlArgc0 + 2
	),
	utter(begin),
	utter(['fid_t  fid = PL_open_foreign_frame();', nl]),
	utter(['term_t t0  = PL_new_term_refs(~d);'+[PlArgc], nl]),
	utter(['qid_t  qid;', nl]),
	utter(['PL_CORBA_Environment plenv;', nl]),
	utter(['static predicate_t pred;', nl]),
	(   Raises == []
	->  true
	;   utter(['term_t except;', nl])
	),
	utter(nl),
	(   debugging(method)
	->  utter(['PL_CORBA_printf("CORBA: Call: ~w::~w()\\n");'+[Itf, Name],
		   nl, nl])
	;   true
	),
	utter(['PL_ILU_find_predicate(&pred, "~w", ~d, "~w", "~w");'+
	       [Name, PlArgc, Itf, ScopeName], nl]),
	(   setting(language, c)
	->  utter(['plenv.CORBA_env = env;', nl])
	;   true
	),
	utter(['plenv.isserver = TRUE;', nl]),
	utter(['plenv.source.predicate = pred;', nl]),

	utter(['PL_put_pointer(t0, this);', nl]),
	write_prolog_arguments(Parms, 1),

	utter(['qid = PL_open_query(NULL, TRUE, pred, t0);', nl, nl]),
	utter(['if ( PL_next_solution(qid) )', nl]),
	utter(begin),
	read_prolog_return_value(Return, PlArgc),
	read_prolog_arguments(Parms, 1),
	utter(['PL_close_query(qid);', nl]),
	utter(['PL_discard_foreign_frame(fid);', nl]),
	return(success, Return),
	utter([end, nl, nl]),
	(   Raises == []
	->  utter(['PL_ILU_error(PL_CORBA_CALLFAILED, &plenv);', nl])
	;   utter(['if ( (except = PL_exception(qid)) )', nl]),
	    utter(begin),
	    utter(['atom_t name; int arity;', nl, nl]),
	    utter(['if ( PL_get_name_arity(except, &name, &arity) )', nl]),
	    utter(begin),
	    except_cases(Raises),
	    utter([end, ' else', nl]),
	    utter(['  PL_ILU_error(PL_CORBA_UNKNOWN_EXCEPTION, &plenv, except);',
		   nl]),
	    utter([end, ' else', nl]),
	    utter(indent(2)),
	    utter(['PL_ILU_error(PL_CORBA_CALLFAILED, &plenv);', nl]),
	    utter(undent)
	),
	utter(nl),
	utter(['PL_close_query(qid);', nl]),
	utter(['PL_discard_foreign_frame(fid);', nl]),
	return(failure, Return),
	utter([end, nl, nl, nl]).
	
return(_Success, void) :- !,
	utter([nl, 'return;', nl]).
return(success, Type) :-
	setting(language, cpp),
	classify_type(Type, interface), !,
	c_type(Type, CType),
	utter([nl, 'return ~w::_narrow(_rval);'+[CType], nl]).
return(success, Type) :-
	setting(language, cpp),
	setting(has_retn, true),
	has_var_type(Type), !,
	utter([nl, 'return _rval._retn();', nl]). % OmniBroker
return(success, _Type) :- !,
	utter([nl, 'return _rval;', nl]).
return(failure, Type) :-
	setting(language, c), !,
	c_type(Type, return, CType),
	utter([nl, 'return (~w)0;'+[CType], nl]).
return(failure, Type) :-		% TBD: use ->nil();
	c_type(Type, return, CType),
	utter([nl, 'return (~w)0;'+[CType], nl]).


%	impl_function(+Scope, +Interface, +Method, -Name)
%
%	determine the name of the function to implement a certain method.

impl_function(Scope, Itf, Method, Name) :-
	setting(language, c), !,
	flatten([server, Scope, Itf, Method], Parts),
	concat_atom(Parts, '_', Name).
impl_function(_Scope, Itf, Method, Name) :-
	setting(namespaces, true),
	setting(impl_class_suffix, Suffix),
	atom_concat(Itf, Suffix, ClassName),
	concat_atom([ClassName, Method], ::, Name).
impl_function(Scope, Itf, Method, Name) :-
	setting(nestedclasses, true),
	flatten([Scope, Itf, Method], Parts),
	concat_atom(Parts, ::, Name).
impl_function(Scope, Itf, Method, Name) :-
	flatten([Scope, Itf], ClassParts),
	concat_atom(ClassParts, '_', ClassName0),
	setting(impl_class_suffix, Suffix),
	atom_concat(ClassName0, Suffix, ClassName),
	concat_atom([ClassName, Method], ::, Name).

%	write_this_arg(+Interface, +Scope)
%
%	Include this in the interface as `this'.  (C-version only)

write_this_arg(Interface, Scope) :-
	setting(language, c), !,
	append(Scope, [Interface], Parts),
	concat_atom(Parts, '_', ThisType),
	utter(['~w this,'+[ThisType], nl, indent]).
write_this_arg(_, _).

%	write_env_arg
%
%	Write (exception) environment argument.  (C-version only)

write_env_arg(Parms) :-
	setting(language, c), !,
	(   Parms \== []
	->  utter([',', nl, indent])
	;   true
	),
        utter(['CORBA_Environment *env']).
write_env_arg(Parms) :-
	setting(orb, orbix), !,
	(   Parms \== []
	->  utter([',', nl, indent])
	;   true			% this is the only argument
	),
        utter(['CORBA(Environment)& env']).
write_env_arg(_).

write_c_function_args([]).
write_c_function_args([H]) :- !,
	utter(['~w'+[H]]).
write_c_function_args([H|T]) :-
	utter(['~w,'+[H], nl, indent]),
	write_c_function_args(T).

c_param(in(Type, Name), Decl) :-
	c_type(Type, in, CType),
	sformat(Decl, '~w ~w', [CType, Name]).
c_param(out(Type, Name), Decl) :-
	c_type(Type, out, CType),
	sformat(Decl, '~w ~w', [CType, Name]).

except_cases([]) :- !,
	utter([nl,
	       '  PL_ILU_error(PL_CORBA_UNKNOWN_EXCEPTION, &plenv, except);',
	       nl]).
except_cases([exception(Name, Scope, _Members)|T]) :-
	concat_atom(Scope, '_', ScopeName),
	concat_atom([ex, ScopeName, Name], '_', FuncName),
	utter(['if ( strcmp(PL_atom_chars(name), "~w") == 0 )'+[Name], nl]),
	utter(['  gen_~w(except, &plenv);'+[FuncName], nl]),
	utter(['else ']),
	except_cases(T).


%	inout() parms!?
%	--> inout(+In, -Out)?

		 /*******************************
		 *	  INPUT ARGUMENTS	*
		 *******************************/

write_prolog_arguments([], _).
write_prolog_arguments([H|T], ArgN) :-
	write_prolog_argument(H, ArgN),
	ArgN1 is ArgN + 1,
	write_prolog_arguments(T, ArgN1).

write_prolog_argument(in(Type, Name), ArgN) :- !,
	sformat(Arg, 't0+~d', [ArgN]),
	indirection(Type, in, Stars),
	write_prolog_arg(in(Name, Type, [pointer(Stars)]), Arg).
write_prolog_argument(_, _).


		 /*******************************
		 *	  OUTPUT ARGUMENTS	*
		 *******************************/

read_prolog_arguments([],  _).
read_prolog_arguments([H|T], ArgN) :-
	read_prolog_argument(H, ArgN),
	ArgN1 is ArgN + 1,
	read_prolog_arguments(T, ArgN1).

read_prolog_argument(out(Type, Name), ArgN) :- !,
	read_prolog_arg(Type, Name, ArgN).
read_prolog_argument(_, _).

read_prolog_arg(Type, Name, ArgN) :-
	setting(language, c),
	indirection(Type, out, Stars),
	sformat(Arg, 't0+~d', [ArgN]),
	(   must_be_allocated(out, Type)
	->  Options = [allocate, pointer(Stars)]
	;   Options = [pointer(Stars)]
	),
	prolog_read_value(Arg,
			  out(Name, Type, Options),
			  '&plenv'), !.
read_prolog_arg(Type, Name, ArgN) :-
	setting(language, cpp),
	sformat(Arg, 't0+~d', [ArgN]),
	(   must_be_allocated(Type)
	->  Options = [allocate, pointer(*)]
	;   Options = []
	),
	prolog_read_value(Arg,
			  out(Name, Type, Options),
			  '&plenv'), !.
read_prolog_arg(Type, _Name, _ArgN) :-
	error(could_not_generate_output_code_for(Type)).


		 /*******************************
		 *	       RETURN		*
		 *******************************/

% Must be tested (also for output args)

read_prolog_return_value(void, _) :- !.
read_prolog_return_value(Type, PlArgc) :-
	setting(language, c),
	c_type(Type, return(var), CType),
	indirection(Type, return(var), Stars),
	utter(['~w _rval;'+[CType], nl, nl]),
	RetArg is PlArgc-1,
	sformat(Arg, 't0+~d', RetArg),
	(   must_be_allocated(Type)
	->  Options = [allocate, pointer(Stars)]
	;   Options = [pointer(Stars)]
	),
	prolog_read_value(Arg,
			  out('_rval', Type, Options),
			  '&plenv'), !.
read_prolog_return_value(Type, PlArgc) :-
	setting(language, cpp),
	c_type(Type, return(var), CType),
	utter(['~w _rval;'+[CType], nl, nl]),
	RetArg is PlArgc-1,
	sformat(Arg, 't0+~d', RetArg),
	(   must_be_allocated(Type)
	->  Options = [allocate, var]
	;   Options = []
	),
	prolog_read_value(Arg,
			  out('_rval', Type, Options),
			  '&plenv'), !.
read_prolog_return_value(Type, _PlArgc) :-
	error(could_not_generate_return_code_for(Type)).


		 /*******************************
		 *	    EXCEPTIONS		*
		 *******************************/

translate_exceptions(Except) :-
	forall(member(E, Except), progress(translate_exception(E))).

translate_exception(exception(Name, Scope, Members)) :-
	setting(language, c), !,
	reset_local_vars,
	concat_atom(Scope, '_', ScopeName),
	utter(['/* Exception "~w" */'+[Name], nl, nl]),
	utter(['static void', nl]),
	concat_atom([ex, ScopeName, Name], '_', FuncName),
	utter(['gen_~w(term_t term, PL_CORBA_Environment *plenv)'+[FuncName],
	       nl]),
	utter(begin),
	  utter(['CORBA_Environment *env = plenv->CORBA_env;', nl]),
	  (   Members == []
	  ->  utter(['env->ptr = NULL;', nl])
	  ;   local_var(arg, ArgVar),
	      utter(['term_t ~w = PL_new_term_ref();'+[ArgVar], nl]),
	      utter(['~w_~w *ctx = ~w_~w__alloc();'
		     +[ScopeName, Name, ScopeName, Name], nl, nl]),
	      utter(['env->ptr = ctx;', nl])
	  ),
	  utter(['env->_major = CORBA_USER_EXCEPTION;', nl]),
	  utter(['env->returnCode = ex_~w_~w;'+[ScopeName, Name], nl]),
	  utter(['env->freeRoutine = NULL;', nl]),
	  prolog_read_except_arg(Members, 1, ArgVar),
	utter([end, nl, nl]).
translate_exception(exception(Name, Scope, Members)) :-
	reset_local_vars,
	concat_atom(Scope, '_', ScopeName),
	utter(['/* Exception "~w" */'+[Name], nl, nl]),
	utter(['static void', nl]),
	concat_atom([ex, ScopeName, Name], '_', FuncName),
	utter(['gen_~w(term_t term, PL_CORBA_Environment *plenv)'+[FuncName],
	       nl]),
	utter(begin),
	(   Members == []
	->  true
	;   local_var(arg, ArgVar),
	    utter(['term_t ~w = PL_new_term_ref();'+[ArgVar], nl])
	),
	named_type(Name, Scope, ExceptClass),
	cpp_except_vars(Members, Vars),
	utter(nl),
	cpp_write_except_vars(Members, 1),
	utter(nl),
	utter(['throw ~w('+[ExceptClass]]),
	utter_except_args(Vars),
	utter([');', nl, end, nl, nl]).

prolog_read_except_arg([], _, _).
prolog_read_except_arg([Name:Type|T], N, ArgVar) :-
	atom_concat('ctx->', Name, Member),
	utter(['PL_get_arg(~d, term, ~w);'+[N, ArgVar], nl]),
	prolog_read_value(ArgVar, out(Member, Type, []), plenv),
	NN is N + 1,
	prolog_read_except_arg(T, NN, ArgVar).

cpp_except_vars([], []).
cpp_except_vars([Name:Type|T0], [Var|T]) :-
	atom_concat('_', Name, Var),
	c_type(Type, var, CType),
	utter(['~w ~w;'+[CType, Var], nl]),
	cpp_except_vars(T0, T).

cpp_write_except_vars([], _).
cpp_write_except_vars([Name:Type|T], N) :-
	atom_concat('_', Name, Var),
	utter(['PL_get_arg(~d, term, arg);'+[N], nl]),
	prolog_read_value(arg, out(Var, Type, []), plenv),
	NN is N + 1,
	cpp_write_except_vars(T, NN).

utter_except_args([]).
utter_except_args([Var]) :- !,
	utter('~w'+[Var]).
utter_except_args([H|T]) :-
	utter('~w, '+[H]),
	utter_except_args(T).


		 /*******************************
		 *       INITIALIZATION		*
		 *******************************/

module_initialiser(Module) :-
	utter(['/* Initialization */', nl, nl]),
	utter(['void', nl]),
	utter(['PL_~w__InitializeServer()'+[Module], nl]),
	utter(begin),
	initialise_module(Module),
	register_foreign_predicates,
	utter([end, nl, nl]),
	(   setting(shared_object, true)
	->  install_function(Module)
	;   true
	).

install_function(Module) :-
	utter(['/* SWI-Prolog install function */', nl, nl]),
	utter(['extern "C" install_t', nl]),
	utter(['install(void)', nl]),
	utter(begin),
	utter(['PL_~w__InitializeServer();'+[Module], nl]),
	utter([end, nl, nl]).

initialise_module(Module) :-
	setting(orb, ilu), !,
	utter(['~w__InitializeServer();'+[Module], nl, nl]).
initialise_module(_).
