/*  $Id: iiop.pl,v 1.1.1.1 2001/01/24 15:20:59 jan Exp $

    Part of XPCE
    Designed and implemented by Anjo Anjewierden and Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/

:- module(iiop,
	  [ iiop_decode_encapsulation/3
	  ]).

		 /*******************************
		 *	   TYPE DECODING	*
		 *******************************/

endianness(1, little) :- !.		% default
endianness(0, big).

% @(Pos,Chars)

iiop_decode_encapsulation([C0|Data], Type, Value) :-
	endianness(C0, Endian),
	phrase(decode(Type, Endian, Value), @(1,Data), @(_,[])).

decode(Type, Endian, Value) -->
	do_decode(Type, Endian, Value), !.
decode(Type, _, _) -->
	{ format('Could not decode for type ~p~n', [Type]),
	  fail
	}.

do_decode(type(Name, _, struct(Members)), Endian, Struct) -->
	{ length(Members, N),
	  functor(Struct, Name, N)
	},
	decode_members(Members, 1, Endian, Struct).
do_decode(type(_Name, _Scope, Type), Endian, Value) -->
	decode(Type, Endian, Value).
do_decode(sequence(Type, _Len), Endian, List) -->
	{   nonvar(List)
	->  length(List, Len)
	;   true
	},
	decode(scalar(unsigned(long)), Endian, Len),
	decode_sequence(Len, Type, Endian, List).
do_decode(scalar(unsigned(long)), Endian, Value) -->
	{ integer(Value), !,
	  C1 is  Value      /\ 0xff,
	  C2 is (Value>> 8) /\ 0xff,
	  C3 is (Value>>16) /\ 0xff,
	  C4 is (Value>>24) /\ 0xff
	},
	align(4),
	(   {Endian == little}
	->  string([C1,C2,C3,C4])
	;   string([C4,C3,C2,C1])
	).
do_decode(scalar(unsigned(long)), big, Value) -->
	align(4),
	string([C1,C2,C3,C4]),
	{ Value is (C1<<24) \/ (C2<<16) \/ (C3<<8) \/ C4
	}.
do_decode(scalar(unsigned(long)), little, Value) -->
	align(4),
	string([C1,C2,C3,C4]),
	{ Value is (C4<<24) \/ (C3<<16) \/ (C2<<8) \/ C1
	}.
do_decode(string(_Len), Endian, Value) -->
	{   atomic(Value)
	->  atom_length(Value, AtomLen),
	    Len is AtomLen+1
	;   true
	},
	decode(scalar(unsigned(long)), Endian, Len),
	decode_string(Len, Value).
do_decode(string(_Len), Endian, Value) -->
	decode(scalar(unsigned(long)), Endian, Len),
	decode_string(Len, Value).
do_decode(scalar(octet), _, Value) -->
	char(Value).

decode_members([], _, _, _) -->
	!.
decode_members([_Name:Type|Members], Arg, Endian, Struct) -->
	{ nonvar(Struct), !,
	  arg(Arg, Struct, Value),
	  NArg is Arg+1
	},
	decode(Type, Endian, Value),
	decode_members(Members, NArg, Endian, Struct).
decode_members([_Name:Type|Members], Arg, Endian, Struct) -->
	decode(Type, Endian, Value),
	{ arg(Arg, Struct, Value),
	  NArg is Arg+1
	},
	decode_members(Members, NArg, Endian, Struct).

decode_sequence(0, _, _, []) -->
	!.
decode_sequence(N, Type, Endian, [H|T]) -->
	decode(Type, Endian, H),
	{NN is N - 1},
	decode_sequence(NN, Type, Endian, T).

decode_string(Len, Atom, @(Old, In), @(New, Out)) :-
	L is Len - 1,			% 0-byte is included in length
	length(S, L),
	append(S, [0|Out], In), !,
	plus(Old, Len, New),
	atom_codes(Atom, S).

align(N, @(Pos,In), @(Pos, In)) :-
	Pos mod N =:= 0, !.
align(N, @(Old,[H|T]), New) :-
	ignore(H=0),
	plus(Old, 1, Tmp),
	align(N, @(Tmp, T), New).

string(List, @(OldPos,In), @(NewPos,Out)) :-
	append(List, Out, In),
	length(List, Len),
	plus(OldPos, Len, NewPos).

char(Char, @(OldPos, [Char|Out]), @(NewPos, Out)) :-
	plus(OldPos, 1, NewPos).
