/*  $Id: ior.pl,v 1.1.1.1 2001/01/24 15:20:59 jan Exp $

    Part of XPCE
    Designed and implemented by Anjo Anjewierden and Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/

:- module(ior,
	  [ decode_ior/2,	% +File, -Info
	    filter_ior/2	% +In, +Out
	  ]).
:- use_module(idl).
:- use_module(iiop).

user:file_search_path(idl, '.').
user:file_search_path(idl, 'pl').

decode_ior(File, Info) :-
	read_ior(File, Bytes),
	ior_type(Type),
	iiop_decode_encapsulation(Bytes, Type, Info).
	

filter_ior(In, Out) :-
	atom_chars(In, InChars),
	phrase(ior(Bytes), InChars, _),
	ior_type(Type),
	iiop_decode_encapsulation(Bytes, Type, 'IOR'(TypeId, Profiles)),
	InetProfile = 'TaggedProfile'(0, _),
	memberchk(InetProfile, Profiles),
	iiop_decode_encapsulation(NewBytes, Type,
				  'IOR'(TypeId, [InetProfile])),
	phrase(ior(NewBytes), OutChars, []),
	atom_chars(Out, OutChars).

ior_type(Type) :-
	absolute_file_name(idl(ior),
			   [ extensions([idl]),
			     access(read)
			   ],
			   IdlFile),
	idl_file_to_term(IdlFile, Def),
	Def = definition(_Dcls,
			 _Itf,
			 [ module('IOR',
				  definition(dcls(_IORConsts,
						  IORTypes,
						  _IORExcept),
					     _Itf2,
					     _SubModules))
			 ]),
	Type = type('IOR', _Scope, _Def),
	memberchk(Type, IORTypes).


vput(X) :-
	between(32, 127, X), !,
	put(X).
vput(X) :-
	format('(~d)', X).

		 /*******************************
		 *	  IOR --> OCTETS	*
		 *******************************/

write_ior(File, Octets) :-
	open(File, write, Fd),
	format(Fd, 'IOR:', []),
	write_octets(Octets, Fd),
	close(Fd).

write_octets([], _).
write_octets([H|T], Fd) :-
	D1 is (H>>4) /\ 0xf,
	D2 is H /\ 0xf,
	hexdigit(C1, D1),
	hexdigit(C2, D2), !,
	put(Fd, C1),
	put(Fd, C2),
	write_octets(T, Fd).

read_ior(File, Octets) :-
	open(File, read, Fd),
	get0(Fd, C0),
	read_file(C0, Fd, Bytes),
	close(Fd),
	phrase(ior(Octets), Bytes, _).
	
read_file(-1, _, []) :- !.
read_file(C0, Fd, [C0|T]) :-
	get0(Fd, C1),
	read_file(C1, Fd, T).


ior(Octets) -->
	"IOR:",
	octets(Octets).

octets([H|T]) -->
	{ nonvar(H),
	  D1 is (H>>4) /\ 0xf,
	  D2 is H /\ 0xf,
	  hexdigit(C1, D1),
	  hexdigit(C2, D2), !
	},
	[C1, C2],
	octets(T).
octets([H|T]) -->
	[C1, C2],
	{ hexdigit(C1, D1),
	  hexdigit(C2, D2),
	  H is 16*D1+D2
	}, !,
	octets(T).
octets([]) -->
	[].


hexdigit(0'0, 0).
hexdigit(0'1, 1).
hexdigit(0'2, 2).
hexdigit(0'3, 3).
hexdigit(0'4, 4).
hexdigit(0'5, 5).
hexdigit(0'6, 6).
hexdigit(0'7, 7).
hexdigit(0'8, 8).
hexdigit(0'9, 9).
hexdigit(0'a, 10).
hexdigit(0'b, 11).
hexdigit(0'c, 12).
hexdigit(0'd, 13).
hexdigit(0'e, 14).
hexdigit(0'f, 15).
hexdigit(0'A, 10).
hexdigit(0'B, 11).
hexdigit(0'C, 12).
hexdigit(0'D, 13).
hexdigit(0'E, 14).
hexdigit(0'F, 15).

