/*  $Id: pl2ilu.c,v 1.3 2001/01/25 15:28:36 jan Exp $

    Designed and implemented by Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "pl2ilu.h"
#include <assert.h>
#ifdef __cplusplus
#include <iostream.h>
#endif
#include <SWI-Stream.h>

#if 0
#define DEBUG(g) g;
#else
#define DEBUG(g) (void)0;
#endif

#ifdef ILU
static int raise_exception(CORBA_Environment *env);
#else
#define raise_exception(s)
#endif

#define FUNCTOR_error2	    PL_new_functor(PL_new_atom("error"), 2)
#define FUNCTOR_type_error2 PL_new_functor(PL_new_atom("type_error"), 2)
static atom_t ATOM_true;		/* "true" */
static atom_t ATOM_false;		/* "false" */

static void
make_constants()
{ if ( !ATOM_true )
  { ATOM_true  = PL_new_atom("true");
    ATOM_false = PL_new_atom("false");
  }
}

		 /*******************************
		 *     ORBIX LOADER SUPPORT    	*
		 *******************************/

#ifdef ORBIX

class loader : CORBA(LoaderClass)
{
  public:
   loader();
   virtual ~loader() {}

   CORBA(Object_ptr) load(const char* itf,
			  const char* marke,
			  CORBA(Boolean),
			  CORBA(Environment)&);
   void		    save(CORBA(Object_ptr),
			 CORBA(saveReason),
			 CORBA(Environment)&);
};
  

loader::loader():CORBA(LoaderClass)(1)
{
}

CORBA(Object_ptr) loader::load(const char* itf,
			       const char* marker,
			       CORBA(Boolean),
			       CORBA(Environment)&)
{ fid_t  fid = PL_open_foreign_frame();
  term_t t0  = PL_new_term_refs(3);
  qid_t  qid;
  predicate_t pred = PL_predicate("corba_loader", 3, "user");
  CORBA(Object_ptr) obj = NULL;

  PL_put_atom_chars(t0,   itf);
  PL_put_atom_chars(t0+1, marker);
  qid = PL_open_query(NULL, TRUE, pred, t0);

  if ( PL_next_solution(qid) )
    PL_get_pointer(t0+2, (void **)&obj);

  PL_close_query(qid);
  PL_discard_foreign_frame(fid);

  if ( obj )
    return obj;

  cerr << "ERROR: loader failed for " << itf << "/" << marker << endl;
  return 0;
}


void loader::save(CORBA(Object_ptr),
		  CORBA(saveReason),
		  CORBA(Environment)&)
{
}

#endif /*ORBIX*/


CORBA_string
PL_ILU_string_dup(const char *s)
{ 
#ifdef ORBIX
  return CORBA(string_dup)(s);
#else
  int len = strlen(s);
  CORBA_string copy = CORBA(string_alloc)(len);
  memcpy(copy, s, len+1);
  return copy;
#endif
}


void
PL_ILU_find_predicate(predicate_t *bin,
		      const char *method,
		      int arity,
		      const char *itf,
		      const char *module)
{ if ( !*bin )
  { char tmp[1024];
    char *s;

    strcpy(tmp, itf);
    s = tmp + strlen(tmp);
    *s++ = '_';
    strcpy(s, method);

    *bin = PL_predicate(tmp, arity, module);
  }
}


int
PL_CORBA_printf(const char *fmt, ...)
{ va_list args;
  int rval;

  va_start(args, fmt);
  rval = Svdprintf(fmt, args);
  va_end(args);

  return rval;
}


int
PL_ILU_error(int error, PL_CORBA_Environment *env, ...)
{ atom_t name;
  int arity;
  module_t module;
  va_list args;
  char predbuf[1024];
  char *pred;

  if ( env->isserver )
  { PL_predicate_info(env->source.predicate, &name, &arity, &module);
    sprintf(predbuf, "%s:%s/%d",
	    PL_atom_chars(PL_module_name(module)),
	    PL_atom_chars(name),
	    arity);
    pred = predbuf;
  } else
  { pred = env->source.predicate_name;
  }

  va_start(args, env);
  switch(error)
  { case PL_CORBA_CALLFAILED:
      PL_CORBA_printf("CORBA server predicate %s failed\n", pred);
      PL_halt(0);			/* should be 1 */
    case PL_CORBA_BADARG:
    { term_t arg = va_arg(args, term_t);
      char *expected = va_arg(args, char *);
      predicate_t print = PL_predicate("print_message", 2, "user");
      fid_t  fid = PL_open_foreign_frame();
      term_t t0 = PL_new_term_refs(2);

      PL_put_atom_chars(t0+0, "error");
      PL_unify_term(t0+1,
		    PL_FUNCTOR, FUNCTOR_error2,
		      PL_FUNCTOR, FUNCTOR_type_error2,
			PL_CHARS, expected,
			PL_TERM, arg,
		      PL_CHARS, pred);

      PL_call_predicate(NULL, TRUE, print, t0);
      PL_discard_foreign_frame(fid);
      PL_halt(0);			/* should be 1 */
    }
    case PL_CORBA_UNKNOWN_EXCEPTION:
    { term_t except = va_arg(args, term_t);
      predicate_t print = PL_predicate("print_message", 2, "user");
      fid_t  fid = PL_open_foreign_frame();
      term_t t0 = PL_new_term_refs(2);

      PL_put_atom_chars(t0+0, "error");
      PL_put_term(t0+1, except);

      PL_call_predicate(NULL, TRUE, print, t0);
      PL_discard_foreign_frame(fid);
      PL_halt(0);			/* should be 1 */
    }
    default:
      PL_CORBA_printf("Unknown error from Prolog CORBA wrapper\n");
      PL_halt(0);			/* should be 1 */
  }
  va_end(args);

  return FALSE;
}



		 /*******************************
		 *	    CONVERSION		*
		 *******************************/

int
PL_CORBA_get_string(term_t t, CORBA_string *s, PL_CORBA_Environment *env)
{ char *tmp;

  if ( PL_get_chars(t, &tmp, CVT_ALL) )
  { *s = PL_ILU_string_dup(tmp);
    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "string");
}


#ifdef IDL_TO_CPP
int
PL_CORBA_get_string(term_t t, CORBA(String_var) *s, PL_CORBA_Environment *env)
{ char *tmp;

  if ( PL_get_chars(t, &tmp, CVT_ALL) )
  {
#ifdef OMNIBROKER
    *s = (const char *)tmp;
#else
    *s = CORBA(string_dup)(tmp);
#endif
    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "string");
}

int
PL_CORBA_get_string(term_t t, CORBA(String_out) s, PL_CORBA_Environment *env)
{ char *tmp;

  if ( PL_get_chars(t, &tmp, CVT_ALL) )
  { s = CORBA(string_dup)(tmp);

    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "string");
}

#ifdef ORBIX
int
PL_CORBA_get_string(term_t t, CORBA(String_mgr) *s, PL_CORBA_Environment *env)
{ char *tmp;

  if ( PL_get_chars(t, &tmp, CVT_ALL) )
  {
#ifdef OMNIBROKER
    *s = (const char *)tmp;
#else
    *s = CORBA(string_dup)(tmp);
#endif
    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "string");
}
#endif/*ORBIX*/
#endif/*IDL_TO_CPP*/


int
PL_CORBA_get_boolean(term_t t, CORBA_boolean *v, PL_CORBA_Environment *env)
{ atom_t a;

  make_constants();

  if ( PL_get_atom(t, &a) )
  { if ( a == ATOM_true )
    { *v = TRUE;
      return TRUE;
    } else if ( a == ATOM_false )
    { *v = FALSE;
      return TRUE;
    }
  }
       
  return PL_ILU_error(PL_CORBA_BADARG, env, t, "boolean");
}


int
PL_CORBA_get_long(term_t t, CORBA_long *v, PL_CORBA_Environment *env)
{ long v0;

  if ( PL_get_long(t, &v0) )
  { *v = (CORBA_long) v0;
    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "long");
}


int
PL_CORBA_get_ulong(term_t t, CORBA_unsigned_long *v, PL_CORBA_Environment *env)
{ long v0;

  if ( PL_get_long(t, &v0) && v0 >= 0 )
  { *v = (CORBA_unsigned_long) v0;
    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "unsigned long");
}

#define MAXOCTET  ((1 << sizeof(char)*8)-1)
#define MAXUSHORT ((1L << sizeof(short)*8)-1)
#undef MAXSHORT
#undef MINSHORT
#define MAXSHORT  ((1L << (sizeof(short)*8-1))-1)
#define MINSHORT  (-(int)MAXSHORT-1)

int
PL_CORBA_get_short(term_t t, short *v, PL_CORBA_Environment *env)
{ int tmp;

  if ( PL_get_integer(t, &tmp) && tmp >= MINSHORT && tmp <= MAXSHORT )
  { *v = (short)tmp;
    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "short");
}


int
PL_CORBA_get_octet(term_t t, unsigned char *v, PL_CORBA_Environment *env)
{ int tmp;

  if ( PL_get_integer(t, &tmp) && tmp >= 0 && tmp <= MAXOCTET )
  { *v = (unsigned char)tmp;
    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "octet");
}


int
PL_CORBA_get_ushort(term_t t, unsigned short *v, PL_CORBA_Environment *env)
{ int tmp;

  if ( PL_get_integer(t, &tmp) && tmp >= 0 && tmp <= MAXUSHORT )
  { *v = (unsigned short)tmp;
    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "unsigned short");
}


int
PL_CORBA_get_float(term_t t, float *v, PL_CORBA_Environment *env)
{ double tmp;

  if ( PL_get_float(t, &tmp) )
  { *v = (float)tmp;
    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "float");
}


int
PL_CORBA_get_double(term_t t, double *v, PL_CORBA_Environment *env)
{ if ( PL_get_float(t, v) )
    return TRUE;

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "double");
}


int
_PL_CORBA_get_object(term_t t, void **v, PL_CORBA_Environment *env)
{ if ( PL_get_pointer(t, v) )
    return TRUE;

  return PL_ILU_error(PL_CORBA_BADARG, env, t, "object");
}


int
PL_CORBA_get_enum(term_t t, CORBA_enum *v,
		  PL_CORBA_enum_def *def,
		  PL_CORBA_Environment *env)
{ char *a;
  const char **list = def->elements;

  if ( PL_get_atom_chars(t, &a) )
  { int n;

    for(n=0; *list; list++, n++)
    { if ( strcmp(*list, a) == 0 )
      { *v = n;
        return TRUE;
      }
    }
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, def->name);
}


void
PL_CORBA_unify_enum(term_t t, CORBA_enum v, PL_CORBA_enum_def *def)
{ const char **list = def->elements;

  PL_unify_atom_chars(t, list[v]);
}


int
PL_CORBA_unify_interface(term_t t, CORBA(Object_ptr) itf)
{ return PL_unify_pointer(t, itf);
}


int
PL_CORBA_get_union(term_t t, const char *uname, term_t type, term_t value,
		   PL_CORBA_Environment *env)
{ atom_t name;
  int arity;

  if ( PL_get_name_arity(t, &name, &arity) && arity == 2 )
  { PL_get_arg(1, t, type);
    PL_get_arg(2, t, value);

    return TRUE;
  }

  return PL_ILU_error(PL_CORBA_BADARG, env, t, uname);
}


void
PL_CORBA_unify_union(term_t t, const char *name, term_t type, term_t value)
{ PL_unify_term(t, PL_FUNCTOR, PL_new_functor(PL_new_atom(name), 2),
		   PL_TERM, type,
		   PL_TERM, value);
}


int
PL_CORBA_unify_boolean(term_t t, CORBA_boolean v)
{ make_constants();

  return PL_unify_atom(t, v ? ATOM_true : ATOM_false);
}


int
PL_CORBA_get_key(term_t t, void **key)
{ long l;

  if ( PL_get_long(t, &l) )
  { *key = (void *) l;
    return TRUE;
  }

  return FALSE;
}


		 /*******************************
		 *         CORBA SUPPORT	*
		 *******************************/

foreign_t
pl_corba_orb(term_t orb)
{
#ifdef ILU
  return PL_unify_atom_chars(orb, "ilu");
#endif
#ifdef ORBIX
  return PL_unify_atom_chars(orb, "orbix");
#endif
#ifdef OMNIBROKER
  return PL_unify_atom_chars(orb, "omnibroker");
#endif
#ifdef ORBACUS
  return PL_unify_atom_chars(orb, "orbacus");
#endif
  return PL_warning("Which ORB???");
}

#ifdef IDL_TO_CPP

static CORBA(ORB_ptr) default_orb;
#ifdef HAVE_BOA
static CORBA(BOA_ptr) default_boa;
#endif
#ifdef HAVE_POA
static PortableServer::POA_ptr default_poa;
//static CORBA(Object_ptr) default_poa;
#endif

CORBA(ORB_ptr)
PL_CORBA_default_server()
{ return default_orb;
}

#ifdef HAVE_POA
PortableServer::POA_ptr
PL_CORBA_default_poa()
{ return default_poa;
}
#endif

#define save_string(s) strcpy((char *)malloc(strlen(s)+1), s)

foreign_t
pl_corba_initialize_server(term_t options, term_t handle)
{ int argc = 0;
  char *argv[100];
  term_t head    = PL_new_term_ref();
  term_t tail    = PL_copy_term_ref(options);
  term_t server  = PL_new_term_ref();
  term_t timeout = PL_new_term_ref();
  char *serverid = NULL;

  DEBUG(cerr << "pl_corba_initialize_server()\n");

  while( PL_get_list(tail, head, tail) )
  { char *s;
    atom_t name;
    int arity;

    if ( PL_get_chars(head, &s, CVT_ALL) )
    { argv[argc++] = save_string(s);
    } else if ( PL_get_name_arity(head, &name, &arity) )
    { int i;
      term_t arg = PL_new_term_ref();
      char *n = (char *)PL_atom_chars(name);

      if ( arity == 1 && strcmp(n, "server") == 0 )
      { PL_get_arg(1, head, server);
	if ( !PL_get_chars(server, &serverid, CVT_ALL) )
	  return FALSE;
      } else if ( arity == 1 && strcmp(n, "timeout") == 0 )
      { PL_get_arg(1, head, timeout);
      } else if ( arity == 1 && strcmp(n, "port") == 0 )
      { term_t port = PL_new_term_ref();
	PL_get_arg(1, head, port);
	int p;
	char pchars[20];

	if ( !PL_get_integer(port, &p) )
	  return FALSE;
	sprintf(pchars, "%d", p);
#ifdef ORBACUS
	argv[argc++] = "-OAport";
	argv[argc++] = save_string(pchars);
#endif
      } else
      { argv[argc++] = n;
	for(i=1; i<=arity; i++)
	{ PL_get_arg(i, head, arg);
	
	  if ( PL_get_chars(arg, &s, CVT_ALL) )
	    argv[argc++] = save_string(s);
	  else
	    return FALSE;
	} 
      }
    }
  }
  if ( !PL_get_nil(tail) )
    return FALSE;

  argv[argc] = NULL;

  CORBA(ORB_ptr) orb;

  try
  {
#ifdef ORBIX
    orb = CORBA(ORB_init)(argc, argv, "Orbix");
#else
    orb = CORBA(ORB_init)(argc, argv);
#endif
  } catch(const CORBA::Exception& ex)
  { cerr << ex << endl;
    return FALSE;
  }

#ifdef ORBIX
  if ( serverid )
    orb->setServerName(serverid);
#endif

#ifdef HAVE_BOA
#ifdef ORBIX
  CORBA(BOA_ptr) boa = orb->BOA_init(argc, argv, "Orbix_BOA");
#else
  CORBA(BOA_ptr) boa = orb->BOA_init(argc, argv);
#endif
  if ( !boa )
    return FALSE;
  default_boa = boa;
#endif

#ifdef HAVE_POA
  CORBA::Object_ptr poaObj = orb->resolve_initial_references("RootPOA");
  default_poa = PortableServer::POA::_narrow(poaObj);
#endif /*HAVE_POA*/

  default_orb = orb;

#ifdef ORBIX
  new loader();
#endif

  return PL_unify_term(handle,
		       PL_FUNCTOR,
		         PL_new_functor(PL_new_atom("CORBA_server"), 4),
		         PL_POINTER, orb,
#ifdef HAVE_BOA
		         PL_POINTER, boa,
#else
		         PL_POINTER, default_poa,
#endif
		         PL_TERM, server,
		         PL_TERM, timeout);
}


foreign_t
pl_corba_initialize_orb(term_t options, term_t handle)
{ int argc = 0;
  char *argv[100];

  argv[argc] = NULL;
#ifdef ORBIX
  CORBA(ORB_ptr) orb = CORBA(ORB_init)(argc, argv, "Orbix");
#else
  CORBA(ORB_ptr) orb = CORBA(ORB_init)(argc, argv);
#endif
  if ( !orb )
    return FALSE;

#ifdef HAVE_BOA
#ifdef ORBIX
  CORBA(BOA_ptr) boa = orb->BOA_init(argc, argv, "Orbix_BOA");
#else
  CORBA(BOA_ptr) boa = orb->BOA_init(argc, argv);
#endif
  if ( !boa )
    return FALSE;
  default_boa = boa;
#else /*HAVE_BOA*/
  void *boa = NULL;
#endif /*HAVE_BOA*/

  default_orb = orb;

#ifdef ORBIX
  orb->setDiagnostics(2);
#endif

  return PL_unify_term(handle,
		       PL_FUNCTOR,
		         PL_new_functor(PL_new_atom("CORBA_orb"), 2),
		         PL_POINTER, orb,
		         PL_POINTER, boa);
}


foreign_t
pl_corba_main_loop(term_t server)
{ 
#ifdef HAVE_BOA
  CORBA(BOA_ptr) boa;
  term_t arg = PL_new_term_ref();

  if ( PL_get_arg(2, server, arg) &&
       PL_get_pointer(arg, (void **)&boa) )
  {
#if defined(OMNIBROKER) || defined(ORBACUS)
    boa->impl_is_ready(CORBA_ImplementationDef::_nil());
#else					/* Orbix */
    CORBA_unsigned_long timeout = CORBA(Orbix).DEFAULT_TIMEOUT;
//  ImplementationDef_ptr server_name = "";
    char *server_name = "";

    if ( PL_get_arg(3, server, arg) )
      PL_get_atom_chars(arg, (char **) &server_name);
    if ( PL_get_arg(4, server, arg) )
    { double t;
      char *s;

      if ( PL_get_float(arg, &t) )
	timeout = (CORBA_unsigned_long) (t * 1000.0);
      else if ( PL_get_atom_chars(arg, &s) && strcmp(s, "default") == 0 )
	timeout = CORBA(Orbix).DEFAULT_TIMEOUT;
      else if ( PL_get_atom_chars(arg, &s) && strcmp(s, "infinite") == 0 )
	timeout = CORBA(Orbix).INFINITE_TIMEOUT;
    }

    try
    { cout << "server name " << server_name << endl;
      boa->impl_is_ready(server_name, timeout);
      //boa->impl_is_ready();
    } catch (CORBA(SystemException) &sysEx )
    { cerr << "Unexpected system exception" << &sysEx << endl;
      exit(1);
    }
#endif

    return TRUE;
  }
#else /*HAVE_BOA*/
#ifdef HAVE_POA

  PortableServer::POAManager_var manager = default_poa->the_POAManager();
  manager->activate();
  PL_CORBA_default_server()->run();

#endif /*HAVE_POA*/
#endif /*HAVE_BOA*/
  
  return FALSE;
}


foreign_t
pl_corba_object_to_string(term_t obj, term_t ior)
{ CORBA(Object_ptr) p;

  if ( PL_get_pointer(obj, (void **)&p) )
  {
#ifdef ORBIX
#ifdef ORBIX_IIOP
    CORBA(String_var) s = CORBA(string_dup)(p->_object_to_string(CORBA_IT_INTEROPERABLE_OR_KIND));
#else
    CORBA(String_var) s = CORBA(string_dup)(p->_object_to_string());
#endif
#else
    CORBA(String_var) s = default_orb->object_to_string(p);
#endif

    return PL_unify_atom_chars(ior, s);
  }

  return FALSE;
}


foreign_t
pl_corba_string_to_object(term_t ior, term_t obj)
{ char *s;

  if ( !default_orb )
    return PL_warning("corba_string_to_object/2: ORB is not initialised");

  if ( PL_get_chars(ior, &s, CVT_ALL) )
  { CORBA(Object_ptr) p = default_orb->string_to_object(s);
    if ( !CORBA(is_nil)(p) )
      return PL_unify_pointer(obj, p);
  }

  return FALSE;
}


foreign_t
pl_corba_duplicate(term_t in, term_t out)
{ CORBA(Object_ptr) p;

  if ( PL_get_pointer(in, (void**)&p) )
  { CORBA(Object)::_duplicate(p);
    
    return PL_unify_pointer(out, p);
  }

  return FALSE;
}


foreign_t
pl_corba_release(term_t in)
{ CORBA(Object_ptr) p;

  if ( PL_get_pointer(in, (void**)&p) )
  { CORBA::release(p);
    
    return TRUE;
  }

  return FALSE;
}

#ifdef HAVE_BOOTMANAGER
#include <OB/BootManager.h>

static
OB::BootManager_ptr
findBootManager()
{ static OB::BootManager_ptr mgr = NULL;

  if ( !mgr )
  { CORBA::Object_ptr obj =
	     default_orb->resolve_initial_references("BootManager");
    mgr = OB::BootManager::_narrow(obj);
  }

  return mgr;
}


int
PL_CORBA_bind(CORBA(Object_ptr) obj, term_t name)
{ char *s;

  if ( PL_get_atom_chars(name, &s) )
  { OB::BootManager_ptr mgr = findBootManager();

    PortableServer::ObjectId_var id = PortableServer::string_to_ObjectId(s);
    mgr->add_binding(id, obj);

    return TRUE;
  }
  
  return FALSE;
}

#endif /*HAVE_BOOTMANAGER*/

#else /*IDL_TO_CPP*/

static ILU_C_Server default_server;

ILU_C_Server
PL_CORBA_default_server()
{ return default_server;
}


foreign_t
pl_corba_initialize_server(term_t options, term_t handle)
{ char *id = NULL;
  char *prot = NULL;
  char *tinfo[10] = { "tcp_0_0", ILU_NIL };
  term_t head = PL_new_term_ref();
  term_t tail = PL_copy_term_ref(options);
  term_t arg  = PL_new_term_ref();
  ILU_C_Server theServer;

  while( PL_get_list(tail, head, tail) )
  { atom_t name;
    int arity;

    if ( PL_get_name_arity(head, &name, &arity) && arity == 1 )
    { const char *n = PL_atom_chars(name);

      PL_get_arg(1, head, arg);

      if ( strcmp(n, "server") == 0 )
      { if ( !PL_get_atom_chars(arg, &id) )
	  return FALSE;
      } else if ( strcmp(n, "protocol") == 0 )
      { if ( !PL_get_atom_chars(arg, &prot) )
	  return FALSE;
      } else if ( strcmp(n, "transport") == 0 )
      { term_t t2 = PL_copy_term_ref(arg);
	term_t h2 = PL_new_term_ref();
	int i=0;

	while(PL_get_list(t2, h2, t2))
	{ char *ti;

	  if ( PL_get_atom_chars(h2, &ti) )
	    tinfo[i++] = ti;
	  else
	    return FALSE;
	}
	if ( PL_get_nil(t2) )
	  tinfo[i++] = ILU_NIL;
	else
	  return FALSE;
      }
    }
  }
  if ( !PL_get_nil(tail) )
    return FALSE;

  { char **v = tinfo;

    PL_CORBA_printf("transport info: ");
    for(; *v; v++)
      PL_CORBA_printf("->%s", *v);
    PL_CORBA_printf("\n");
  }

  theServer = ILU_C_InitializeServer(id,	/* id */
				     NULL,      /* object table */
				     prot,	/* protocol */
				     tinfo,	/* transport */
				     NULL,	/* passport */
				     ilu_TRUE);	/* use def port */
  if ( theServer )
  { if ( !default_server )
      default_server = theServer;

    return PL_unify_pointer(handle, theServer);
  }

  return FALSE;
}


foreign_t
pl_corba_main_loop(term_t server)
{ ILU_C_Run();

  return TRUE;
}


foreign_t
pl_corba_object_to_string(term_t obj, term_t string)
{ void *ptr;

  if ( PL_get_pointer(obj, &ptr) )
  { CORBA_Environment env;
    char *s = CORBA_ORB_object_to_string(ILU_C_ORB,
					 (CORBA_Object) ptr,
					 &env);

/*  PL_CORBA_printf("SBH=\"%s\"\n", ILU_C_SBHOfObject(ptr)); */

    if ( s )
      return PL_unify_atom_chars(string, s);
  }

  return FALSE;
}


foreign_t
pl_corba_duplicate(term_t in, term_t out)
{ CORBA_Object from;

  if ( PL_get_pointer(in, (void **)&from) )
  { CORBA_Environment env;
    CORBA_Object to;

    if ( (to = CORBA_Object_duplicate(from, &env)) )
      return PL_unify_pointer(out, to);
  }

  return FALSE;
}


foreign_t
pl_corba_release(term_t in)
{ CORBA(Object_ptr) p;

  if ( PL_get_pointer(in, (void**)&p) )
  { /*TBD*/
    
    return TRUE;
  }

  return FALSE;
}


foreign_t
pl_corba_initialize_orb(term_t options, term_t handle)
{ return TRUE;
}


		 /*******************************
		 *        CLIENT SUPPORT	*
		 *******************************/

foreign_t
pl_corba_string_to_object(term_t ior, term_t obj)
{ char *s;

  if ( PL_get_atom_chars(ior, &s) )
  { CORBA_Environment env;
    void *self = CORBA_ORB_string_to_object(ILU_C_ORB, s, &env);

    if ( self )
      return PL_unify_pointer(obj, self);

    raise_exception(&env);
  }

  return FALSE;
}


static int
raise_exception(CORBA_Environment *env)
{ switch(env->_major)
  { case CORBA_NO_EXCEPTION:
      return FALSE;
    case CORBA_SYSTEM_EXCEPTION:
    { char *id = CORBA_exception_id(env);
      PL_CORBA_exception(PL_CORBA_SYSTEM_EXCEPTION, id);
      CORBA_exception_free(env);

      return TRUE;
    }
    default:
      assert(0);
  }
}
#endif


void
PL_CORBA_exception(int type, ...)
{ va_list args;

  va_start(args, type);

  switch(type)
  { case PL_CORBA_SYSTEM_EXCEPTION:
    { char *id = va_arg(args, char *);
      term_t et = PL_new_term_ref();

      PL_unify_term(et,
		    PL_FUNCTOR, PL_new_functor(PL_new_atom("corba_system_exception"), 1),
		    PL_CHARS, id);

      PL_throw(et);
      break;
    }
    case PL_CORBA_UNKNOWN_USER_EXCEPTION:
    { char *id = va_arg(args, char *);
      term_t et = PL_new_term_ref();

      PL_unify_term(et,
		    PL_FUNCTOR, PL_new_functor(PL_new_atom("corba_unknown_user_exception"), 1),
		    PL_CHARS, id);

      PL_throw(et);
      break;
    }
    case PL_CORBA_USER_EXCEPTION:
    { term_t t = va_arg(args, term_t);
      term_t et = PL_new_term_ref();

      PL_unify_term(et,
		    PL_FUNCTOR, PL_new_functor(PL_new_atom("corba_user_exception"), 1),
		    PL_TERM, t);

      PL_throw(t);
      break;
    }
    default:
      assert(0);
  }

  va_end(args);
}


