/*  $Id: pl2ilu.h,v 1.2 2001/01/25 15:28:36 jan Exp $

    Designed and implemented by Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/

#ifndef PL2ILU_H_INCLUDED
#define PL2ILU_H_INCLUDED

#include <SWI-Prolog.h>
#include <string.h>
#include <assert.h>

#ifndef __cplusplus
#define ILU 1
#endif

#ifdef ILU
#include <iluchdrs.h>
#define IDL_TO_C 1
#else
#if defined(OMNIBROKER) || defined(ORBACUS)
#include <OB/CORBA.h>
#define HAVE_POA 1			/* use POA */
#define HAVE_BOOTMANAGER 1		/* use bootmanager to publish */
#else
#include <CORBA.h>
#define HAVE_BOA 1			/* use BOA */
#endif
#define IDL_TO_CPP 1
#endif

#ifndef ORBIX
#ifdef ORBACUS
#define CORBA(id) CORBA::id
#else
#define CORBA(id) CORBA_ ## id
#endif
#endif

#ifdef IDL_TO_CPP
typedef char *		CORBA_string;
typedef CORBA(Long)	CORBA_long;
typedef CORBA(ULong)	CORBA_unsigned_long;
typedef CORBA(Short)	CORBA_short;
typedef CORBA(UShort)	CORBA_unsigned_short;
typedef CORBA(ULong)	CORBA_enum;
typedef CORBA(Boolean)	CORBA_boolean;
#endif

#ifndef CORBA_String
#define CORBA_String    char *
#endif

#ifdef IDL_TO_C
typedef CORBA(Object)	CORBA(Object_ptr);
#endif

#ifndef NULL
#define NULL (0)
#endif


#define PL_CORBA_BADARG            1	/* argument translation failed */
#define PL_CORBA_CALLFAILED        2	/* call to server predicate failed */
#define PL_CORBA_UNKNOWN_EXCEPTION 3	/* thrown bad exception */

#define PL_CORBA_SYSTEM_EXCEPTION  0
#define PL_CORBA_UNKNOWN_USER_EXCEPTION 1
#define PL_CORBA_USER_EXCEPTION    2

typedef struct
{
#ifdef IDL_TO_C
  CORBA_Environment *CORBA_env;
#endif
  int		     isserver;		/* server code */
  union
  { predicate_t	     predicate;		/* server */
    char *	     predicate_name;	/* client */
  } source;
} PL_CORBA_Environment;

typedef struct
{ char *name;				/* name of the enum */
  const char **elements;
} PL_CORBA_enum_def;


#define FUNCTOR_corba2 _FUNCTOR_corba2()

extern CORBA_string PL_ILU_string_dup(const char *s);
extern void	    PL_ILU_find_predicate(predicate_t *bin,
					  const char *method,
					  int arity,
					  const char *itf,
					  const char *module);
extern int	PL_CORBA_printf(const char *fmt, ...);
extern int      PL_ILU_error(int error,
			     PL_CORBA_Environment *env, ...);
PL_CORBA_Environment	*PL_CORBA_get_environment(term_t context);
extern functor_t	_FUNCTOR_corba2(void);

extern int	PL_CORBA_get_string(term_t t,
				    CORBA_string *s,
				    PL_CORBA_Environment *env);
#ifdef IDL_TO_CPP
extern int	PL_CORBA_get_string(term_t t,
				    CORBA(String_var)* s,
				    PL_CORBA_Environment *env);
					/* why not CORBA(String)? */
extern int	PL_CORBA_get_string(term_t t,
				    CORBA(String_out) s,
				    PL_CORBA_Environment *env);
#ifdef ORBIX
extern int	PL_CORBA_get_string(term_t t,
				    CORBA(String_mgr) *s,
				    PL_CORBA_Environment *env);
#endif
#endif
extern int	PL_CORBA_get_boolean(term_t t0,
				     CORBA_boolean *v,
				     PL_CORBA_Environment *env);
extern int	PL_CORBA_get_long(term_t t0,
				  CORBA_long *v,
				  PL_CORBA_Environment *env);
extern int	PL_CORBA_get_short(term_t t0,
				   short *v,
				   PL_CORBA_Environment *env);
extern int	PL_CORBA_get_octet(term_t t0,
				   unsigned char *v,
				   PL_CORBA_Environment *env);
extern int	PL_CORBA_get_ulong(term_t t0,
				   CORBA_unsigned_long *v,
				   PL_CORBA_Environment *env);
extern int	PL_CORBA_get_ushort(term_t t0,
				    unsigned short *v,
				    PL_CORBA_Environment *env);
extern int	PL_CORBA_get_float(term_t t0,
				   float *v,
				   PL_CORBA_Environment *env);
extern int	PL_CORBA_get_double(term_t t0,
				    double *v,
				    PL_CORBA_Environment *env);
#define PL_CORBA_get_object(t, v, env) \
	_PL_CORBA_get_object((t), (void **)(v), (env))
extern int	_PL_CORBA_get_object(term_t t,
				    void **v,
				    PL_CORBA_Environment *env);
extern int	PL_CORBA_get_enum(term_t t0,
				  CORBA_enum *v,
				  PL_CORBA_enum_def *def,
				  PL_CORBA_Environment *env);
extern void	PL_CORBA_unify_enum(term_t t0,
				    CORBA_enum v,
				    PL_CORBA_enum_def *def);
extern int	PL_CORBA_unify_interface(term_t t,
					 CORBA(Object_ptr));
extern int	PL_CORBA_unify_boolean(term_t t, CORBA_boolean v);
extern void	PL_CORBA_unify_union(term_t t,
				     const char *name,
				     term_t type,
				     term_t value);
extern int	PL_CORBA_get_union(term_t t, const char *name,
				   term_t type, term_t value,
				   PL_CORBA_Environment *env);

extern int	PL_CORBA_get_key(term_t t, void **key);

		 /*******************************
		 *	   CORBA CONTROL	*
		 *******************************/

#ifdef ILU
extern ILU_C_Server PL_CORBA_default_server(void);
#else
extern CORBA(ORB_ptr) PL_CORBA_default_server(void);
#ifdef HAVE_POA
extern PortableServer::POA_ptr PL_CORBA_default_poa(void);
#endif
#endif
extern foreign_t pl_corba_orb(term_t orb);
extern foreign_t pl_corba_initialize_server(term_t options, term_t server);
extern foreign_t pl_corba_initialize_orb(term_t options, term_t orb);
extern foreign_t pl_corba_main_loop(term_t server);
extern foreign_t pl_corba_object_to_string(term_t obj, term_t ior);
extern foreign_t pl_corba_string_to_object(term_t ior, term_t obj);
extern foreign_t pl_corba_duplicate(term_t in, term_t out);
extern foreign_t pl_corba_release(term_t obj);

#ifdef HAVE_BOOTMANAGER
extern int PL_CORBA_bind(CORBA(Object_ptr) obj, term_t name);
#endif

		 /*******************************
		 *	      CLIENT		*
		 *******************************/

extern void	PL_CORBA_exception(int type, ...);

#endif /*PL2ILU_H_INCLUDED*/
