/*  $Id: plskel.pl,v 1.1.1.1 2001/01/24 15:20:59 jan Exp $

    Designed and implemented by Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/

:- module(plskel,
	  [ pl_method_skeleton/3,	% +Itf, +Scope, +Method
	    pl_interface_header/1,	% +Itf
	    pl_header/1			% +Module
	  ]).

pl_header(Module) :-
	format(plout, '/*  Prolog code for server CORBA interface~n', []),
	format(plout, '*/~n~n~n', []),
	format(plout, ':- module(~q, []).~n~n~n', [Module]).

pl_method_skeleton(Itf, _Scope,
		   method(Name, Parms, Return, Raises, _Context)) :-
	pptype(Return, IDLReturn),
	format(plout, '/* ~w ~w(', [IDLReturn, Name]),
	pl_arg_comments(Parms),
	format(plout, ')', []),
	pl_raises_comment(Raises),
	format(plout, ' */~n~n', []),
	concat_atom([Itf, Name], '_', Pred),
	format(plout, '~q(Self', [Pred]),
	pl_arg_skeletons(Parms),
	(   Return \== void
	->  format(plout, ', Return', [])
	;   true
	),
	format(plout, ') :-~n~t~8|fail.~n~n~n', []).

%	pl_raises_comment(Exceptions)
%	Document the exceptions that can be raised by a method.

pl_raises_comment([]) :- !.
pl_raises_comment(Exceptions) :-
	format(plout, '~n~t~8|raises(', []),
	pl_except_list(Exceptions),
	format(plout, ')', []).

pl_except_list([E]) :- !,
	pl_except(E).
pl_except_list([H|T]) :-
	pl_except(H),
	format(plout, ',', []),
	pl_except_list(T).

pl_except(exception(Name, _Scope, _Members)) :-
	format(plout, '~w', [Name]).

%	Document the method arguments

pl_arg_comments([]).
pl_arg_comments([X]) :- !,
	pl_arg_comment(X).
pl_arg_comments([H|T]) :- !,
	pl_arg_comment(H),
	format(plout, ', ', []),
	pl_arg_comments(T).

pl_arg_comment(in(Type, Name)) :-
	prolog_argument(Name, VarName),
	pptype(Type, IdlType),
	format(plout, 'in ~w ~w', [IdlType, VarName]).
pl_arg_comment(out(Type, Name)) :-
	prolog_argument(Name, VarName),
	pptype(Type, IdlType),
	format(plout, 'out ~w ~w', [IdlType, VarName]).

%	Document a type

pptype(type(Name, _, _), Name) :- !.
pptype(scalar(unsigned(What)), Type) :- !,
	concat_atom([unsigned, ' ', What], Type).
pptype(scalar(What), What) :- !.
pptype(string(unbound), string) :- !.
pptype(string(N), S) :- !,
	sformat(S, 'string<~d>', N).
pptype(X, X).

%	Generate the Prolog method argument template

pl_arg_skeletons([]).
pl_arg_skeletons([H|T]) :- !,
	format(plout, ', ', []),
	pl_arg_skeleton(H),
	pl_arg_skeletons(T).

pl_arg_skeleton(in(_Type, Name)) :-
	prolog_argument(Name, VarName),
	format(plout, '~w', [VarName]).
pl_arg_skeleton(out(_Type, Name)) :-
	prolog_argument(Name, VarName),
	format(plout, '~w', [VarName]).

%	prolog_argument(+IdlIdentifier, -PrologVarName)

prolog_argument(IdlName, VarName) :-
	atom_codes(IdlName, Chars0),
	capitalise(Chars0, +, Chars),
	atom_codes(VarName, Chars).

%	pl_interface_header(+Itf)
%
%	Write a header for the interface.

pl_interface_header(Itf) :-
	pl_section_header(Itf).

%	pl_section_header(+Title)
%
%	Write a nice looking header


pl_section_header(Title) :-
	upcase_atom(Title, UpperTitle),
	format(plout, '~N~t~16|/~`*t~48|~n', []),
	format(plout, '~t~16|*~t~w~t~47|*~n', [UpperTitle]),
	format(plout, '~t~16|~`*t~47|/~n~n~n', []).

		 /*******************************
		 *	       UTIL		*
		 *******************************/

capitalise([], _, []).
capitalise([0'_,C0|T0], _, [C1|T1]) :- !,
	upcase(C0, C1),
	capitalise(T0, -, T1).
capitalise([C0|T0], +, [C1|T1]) :- !,
	upcase(C0, C1),
	capitalise(T0, -, T1).
capitalise([C|T0], _, [C|T1]) :-
	capitalise(T0, -, T1).

upcase_atom(In, Out) :-
	atom_codes(In, Chars0),
	upcase_chars(Chars0, Chars),
	atom_codes(Out, Chars).

upcase_chars([], []).
upcase_chars([H0|T0], [H|T]) :-
	upcase(H0, H),
	upcase_chars(T0, T).

upcase(L, U) :-
	between(0'a, 0'z, L), !,
	U is L + 0'A - 0'a.
upcase(C, C).

