/*  $Id: util.pl,v 1.1.1.1 2001/01/24 15:20:59 jan Exp $

    Part of CORBA <-> Prolog interface
    Designed and implemented by Anjo Anjewierden and Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/

:- module(idl2pl_util,
	  [ progress/1,
	    verbose/1,
	    error/1,

	    push/2,			% +Name, +Value
	    pop/1,			% +Name
	    setting/2,			% +Name, -Value
	    debugging/1,		% +Subject
	    reset_settings/0
	  ]).

:- meta_predicate(progress(:)).

progress(Goal) :-
	verbose(start(Goal)),
	(   Goal
	->  verbose(true(Goal))
	;   verbose(false(Goal)),
	    fail
	).

verbose(Term) :-
	setting(verbose, true), !,
	(   message_to_format(Term, Fmt, Args)
	->  format(user_error, Fmt, Args)
	).
verbose(_).

	
error(Term) :-
	message_to_format(Term, Fmt, Args), !,
	format(user_error, '~NERROR: ', []),
	format(user_error, Fmt, Args).

idl_message(start(Goal)) -->
	goal_message(Goal),
	[ ' ... ' ].
idl_message(true(_Goal)) -->
	[ 'ok', nl ].
idl_message(false(_Goal)) -->
	[ 'FAILED!', nl ].
idl_message(Term) -->
	[ '~p'-[Term] ].

goal_message(parse_idl(_File, _Tokens)) -->
	[ parsing ].
goal_message(cwrapper_for_module(module(Module, _Defs))) -->
	[ 'making wrapper for module "~w"'-[Module] ].
goal_message(translate_exceptions(_)) -->
	[ '~N~t~2|exceptions'-[] ].
goal_message(translate_exception(exception(Name, _, _))) -->
	[ '~N~t~4|exception "~w"'-[Name] ].
goal_message(translate_interface(interface(Name, _, _, _, _))) -->
	[ '~t~2|interface "~w"'-[Name] ].
goal_message(function_wrapper(_Interface, _Scope,
			      method(Name, _Parms, _Return, _Excp, _Ctx))) -->
	[ '~N~t~4|method "~w"'-[Name] ].
goal_message(module_initialiser(_)) -->
	[ '~t~2|initializer'-[] ].
goal_message(Goal) -->
	{ functor(Goal, Name, Arity) },
	[ 'running ~w/~d'-[Name, Arity] ].




%	message_to_string(+Term, -String)
%
%	Translate an error term into a string

message_to_format(Term, Fmt, Args) :-
        idl_message(Term, Actions, []), !,
        actions_to_format(Actions, Fmt, Args).

actions_to_format([], '', []) :- !.
actions_to_format([nl|T], Fmt, Args) :- !,
	actions_to_format(T, Fmt0, Args),
	atom_concat('~n', Fmt0, Fmt).
actions_to_format([Fmt0-Args0|Tail], Fmt, Args) :- !,
        actions_to_format(Tail, Fmt1, Args1),
        atom_concat(Fmt0, Fmt1, Fmt),
        append(Args0, Args1, Args).
actions_to_format([Term|Tail], Fmt, Args) :-
	atomic(Term), !,
        actions_to_format(Tail, Fmt1, Args),
	atom_concat(Term, Fmt1, Fmt).
actions_to_format([Term|Tail], Fmt, Args) :-
        actions_to_format(Tail, Fmt1, Args1),
        atom_concat('~w', Fmt1, Fmt),
        append([Term], Args1, Args).

		 /*******************************
		 *	       CONTEXT		*
		 *******************************/
:- dynamic
	store/2.			% What, Value

default(indent, 0).

push(What, Value) :-
	asserta(store(What, Value)).

pop(What) :-
	retract(store(What, _)), !.

setting(What, Value) :-
	store(What, X), !,
	Value = X.
setting(What, Value) :-
	default(What, Value).

debugging(What) :-
	store(debugging, What), !.

reset_settings :-
	retractall(store(_, _)).
