/*  Prolog code for server CORBA interface
*/


:- module(test, []).
:- use_module(tests).

term_expansion(expand_tests, Tests) :-
	findall(Test, make_test(Test), Tests).

make_test((Head :- Body)) :-
	test_interface(input),
	test_data(DataName, Term),
	concat_atom([input, t, DataName], '_', Pred),
	Head =.. [Pred, _Self, Input],
	Body = (   Input == Term
	       ->  format('~N~w ... ok~n', [Pred])
	       ;   format('~N~w ... Wrong Input!~n', [Pred])
	       ).
make_test(Head) :-
	test_interface(Name),
	memberchk(Name, [answer, output]),
	test_data(DataName, Term),
	concat_atom([Name, t, DataName], '_', Pred),
	Head =.. [Pred, _Self, Term].
make_test((Head :- throw(Except))) :-
	test_interface(except),
	test_data(DataName, Term),
	concat_atom([except, t, DataName], '_', Pred),
	concat_atom([t_except, DataName], '_', ExceptName),
	Except =.. [ExceptName, Term],
	Head =.. [Pred, _Self].
		   

		 /*******************************
		 *	       ...		*
		 *******************************/

expand_tests.


                /*******************************
                *           FACTORY            *
                *******************************/


/* input make_input() */

fabriek_make_input(_Self, Return) :-
	input_create_true(input, Return0),
	corba_duplicate(Return0, Return).


/* answer make_answer() */

fabriek_make_answer(_Self, Return) :-
	answer_create_true(answer, Return0),
	corba_duplicate(Return0, Return).


/* output make_output() */

fabriek_make_output(_Self, Return) :-
	output_create_true(output, Return0),
	corba_duplicate(Return0, Return).


/* except make_except() */

fabriek_make_except(_Self, Return) :-
	except_create_true(except, Return0),
	corba_duplicate(Return0, Return).


